from collections import OrderedDict
from datetime import datetime, date
from unittest import TestCase
from xml.etree.ElementTree import ElementTree

import numpy as np

from esa_climate_toolbox.util.misc import encode_url_path
from esa_climate_toolbox.util.misc import new_indexed_name
from esa_climate_toolbox.util.misc import object_to_qualified_name
from esa_climate_toolbox.util.misc import qualified_name_to_object
from esa_climate_toolbox.util.misc import to_datetime, to_datetime_range
from esa_climate_toolbox.util.misc import to_json
from esa_climate_toolbox.util.misc import to_list
from esa_climate_toolbox.util.misc import to_scalar
from esa_climate_toolbox.util.misc import to_str_constant, is_str_constant
from esa_climate_toolbox.util.undefined import UNDEFINED


class UtilTest(TestCase):

    def test_object_to_qualified_name(self):
        self.assertEqual(object_to_qualified_name(float),
                         'float')
        self.assertEqual(object_to_qualified_name(np.dtype('float64')),
                         'float64')
        self.assertEqual(object_to_qualified_name(TestCase),
                         'unittest.case.TestCase')
        self.assertEqual(object_to_qualified_name(ElementTree),
                         'xml.etree.ElementTree.ElementTree')
        self.assertEqual(object_to_qualified_name({}, fail=False),
                         '{}')
        with self.assertRaisesRegex(ValueError, "missing attribute '__name__'"):
            object_to_qualified_name({}, fail=True)

    def test_qualified_name_to_object(self):
        self.assertIs(qualified_name_to_object('float'), float)
        self.assertIs(qualified_name_to_object('builtins.float'), float)
        self.assertIs(qualified_name_to_object('unittest.case.TestCase'),
                      TestCase)
        self.assertIs(
            qualified_name_to_object('xml.etree.ElementTree.ElementTree'),
            ElementTree
        )
        with self.assertRaisesRegex(ImportError, "No module named 'numpi'"):
            qualified_name_to_object('numpi.ndarray')
        with self.assertRaisesRegex(
                AttributeError, "module 'builtins' has no attribute 'flaot'"):
            qualified_name_to_object('flaot')

    def test_encode_path(self):
        self.assertEqual(encode_url_path(
            '/ws/init', query_args=OrderedDict([
                ('base_path', '/home/workpaces'), ('description', 'Hi there!')
            ])),
            '/ws/init?base_path=%2Fhome%2Fworkpaces&description=Hi+there%21'
        )
        self.assertEqual(encode_url_path(
            '/ws/init', query_args=OrderedDict([
                ('base_path', 'C:\\Users\\Norman\\workpaces'),
                ('description', 'Hi there!')])),
            '/ws/init?base_path=C%3A%5CUsers%5CNorman%5Cworkpaces&'
            'description=Hi+there%21')

        self.assertEqual(encode_url_path(
            '/ws/get/{base_path}',
            path_args=dict(base_path='/home/norman/workpaces')),
            '/ws/get/%2Fhome%2Fnorman%2Fworkpaces'
        )
        self.assertEqual(encode_url_path(
            '/ws/get/{base_path}',
            path_args=dict(base_path='C:\\Users\\Norman\\workpaces')),
            '/ws/get/C%3A%5CUsers%5CNorman%5Cworkpaces'
        )

    def test_to_datetime(self):
        dt = to_datetime('1998-11-20 10:14:08', default=None)
        self.assertIsInstance(dt, datetime)
        self.assertEqual(datetime(1998, 11, 20, 10, 14, 8), dt)

        dt = to_datetime('2001-01-01', default=None)
        self.assertIsInstance(dt, datetime)
        self.assertEqual(datetime(2001, 1, 1), dt)

        dt = to_datetime('2002-03', default=None)
        self.assertIsInstance(dt, datetime)
        self.assertEqual(datetime(2002, 3, 1), dt)

        dt = to_datetime('2003', default=None)
        self.assertIsInstance(dt, datetime)
        self.assertEqual(datetime(2003, 1, 1), dt)

        dt = to_datetime('1998-11-20 10:14:08', upper_bound=True, default=None)
        self.assertIsInstance(dt, datetime)
        self.assertEqual(datetime(1998, 11, 20, 10, 14, 8), dt)

        dt = to_datetime('2001-01-01', upper_bound=True, default=None)
        self.assertIsInstance(dt, datetime)
        self.assertEqual(datetime(2001, 1, 1, 23, 59, 59), dt)

        dt = to_datetime('2002-03', upper_bound=True, default=None)
        self.assertIsInstance(dt, datetime)
        self.assertEqual(datetime(2002, 3, 28, 23, 59, 59), dt)

        dt = to_datetime('2003', upper_bound=True, default=None)
        self.assertIsInstance(dt, datetime)
        self.assertEqual(datetime(2003, 12, 31, 23, 59, 59), dt)

        dt = to_datetime('2001-01-01 2:3:5', default=None)
        self.assertIsInstance(dt, datetime)
        self.assertEqual(datetime(2001, 1, 1, 2, 3, 5), dt)

        dt = to_datetime(datetime(2001, 1, 1), default=None)
        self.assertIsInstance(dt, datetime)
        self.assertEqual(datetime(2001, 1, 1), dt)

        dt = to_datetime(date(2012, 4, 20), default=None)
        self.assertIsInstance(dt, datetime)
        self.assertEqual(datetime(2012, 4, 20, 12), dt)

        dt = to_datetime(None, default=datetime(2001, 1, 1))
        self.assertIsInstance(dt, datetime)
        self.assertEqual(datetime(2001, 1, 1), dt)

        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            to_datetime(1, default=None)
        with self.assertRaises(ValueError):
            to_datetime("42", default=None)

    def test_to_datetime_range(self):
        dtr = to_datetime_range(None, '')
        self.assertIsNone(dtr)

        dtr = to_datetime_range('2008', None)
        self.assertEqual(dtr[0], datetime(2008, 1, 1))
        self.assertEqual(dtr[1], datetime(2008, 12, 30, 23, 59, 59))

        dtr = to_datetime_range('2008-08', None)
        self.assertEqual(dtr[0], datetime(2008, 8, 1))
        self.assertEqual(dtr[1], datetime(2008, 8, 28, 23, 59, 59))

        dtr = to_datetime_range('2008-10-10', None)
        self.assertEqual(dtr[0], datetime(2008, 10, 10, 0, 0))
        self.assertEqual(dtr[1], datetime(2008, 10, 10, 23, 59, 59))

        with self.assertRaises(ValueError):
            to_datetime_range("211", "2012")

    def test_new_indexed_name(self):
        with self.assertRaises(ValueError) as cm:
            new_indexed_name(['var_1', 'var_2'], 'ds_{id}')
        self.assertEqual(str(cm.exception), 'pattern must contain "{index}"')

        with self.assertRaises(ValueError) as cm:
            new_indexed_name(['var_1', 'var_2'], '{index}_plot')
        self.assertEqual(
            str(cm.exception), 'pattern does not yield a valid name'
        )

        self.assertEqual(
            new_indexed_name(['var_1', 'res_4', 'var_2'], 'ds_{index}'),
            'ds_1'
        )
        self.assertEqual(
            new_indexed_name(['var_1', 'res_4', 'var_2'], 'var_{index}'),
            'var_3'
        )
        self.assertEqual(
            new_indexed_name(['var_1', 'res_4', 'var_2'], 'res_{index}'),
            'res_5'
        )
        self.assertEqual(
            new_indexed_name(['var_1', 'res_4', 'var_2_subs'],
                             'var_{index}_subs'),
            'var_3_subs'
        )
        self.assertEqual(
            new_indexed_name(['var_1', 'res_4', 'var_4_3'],
                             'var_{index}_{index}'),
            'var_5_5'
        )

        self.assertEqual(new_indexed_name([], 'ds_{index}'), 'ds_1')
        self.assertEqual(new_indexed_name(['ds5'], 'ds_{index}'), 'ds_1')
        self.assertEqual(new_indexed_name(['ds_005'], 'ds_{index}'), 'ds_6')


class ToListTest(TestCase):

    def test_none_and_empty(self):
        self.assertEqual(to_list(None), None)
        self.assertEqual(to_list([]), [])

    def test_str(self):
        self.assertEqual(to_list('a'), ['a'])
        self.assertEqual(to_list('a, b, c'), ['a', 'b', 'c'])
        self.assertEqual(to_list(['a', 'b', 'c']), ['a', 'b', 'c'])
        self.assertEqual(to_list(('a', 'b', 'c')), ['a', 'b', 'c'])
        self.assertEqual(to_list([1, 2, 3]), ['1', '2', '3'])

    def test_int(self):
        self.assertEqual(to_list(1, dtype=int), [1])
        self.assertEqual(to_list('1, 2, 3', dtype=int), [1, 2, 3])
        self.assertEqual(to_list([1, 2, 3], dtype=int), [1, 2, 3])
        self.assertEqual(to_list((1, 2, 3), dtype=int), [1, 2, 3])
        self.assertEqual(to_list(['1', '2', '3'], dtype=int), [1, 2, 3])


class ToJsonTest(TestCase):

    def test_none_and_empty(self):
        self.assertEqual(to_json(None), None)
        self.assertEqual(to_json([]), [])
        self.assertEqual(to_json({}), {})
        self.assertEqual(to_json(''), '')

    def test_numpy(self):
        self.assertEqual(to_json(np.array([])),
                         [])
        self.assertEqual(to_json(np.array([1])),
                         [1])
        self.assertEqual(to_json(np.array([1, 2, 3])),
                         [1, 2, 3])
        self.assertEqual(to_json(np.array([[1, 2], [3, 4]])),
                         [[1, 2], [3, 4]])
        self.assertEqual(to_json(np.array([1, 2, 3])[2]),
                         3)
        self.assertEqual(to_json(np.array([np.datetime64('2005-02-21'),
                                           np.datetime64('2005-02-23'),
                                           np.datetime64('2005-02-25')])),
                         ['2005-02-21',
                          '2005-02-23',
                          '2005-02-25'])
        self.assertEqual(
            to_json(np.array([np.datetime64('2005-02-21')])), ['2005-02-21']
        )
        self.assertEqual(to_json(np.ndarray), 'numpy.ndarray')

    def test_types(self):
        self.assertEqual(to_json(str), "str")
        self.assertEqual(to_json(OrderedDict), "collections.OrderedDict")

    def test_scalar_values(self):
        self.assertEqual(to_json(True), True)
        self.assertEqual(to_json(False), False)
        self.assertEqual(to_json('ohoh'), 'ohoh')
        self.assertEqual(to_json(234), 234)
        self.assertEqual(to_json(4.6), 4.6)
        self.assertEqual(
            to_json(np.array(np.datetime64('2005-02-21'))), '2005-02-21'
        )

    def test_composite_values(self):
        self.assertEqual(to_json([1, 'no!', 3]), [1, 'no!', 3])
        self.assertEqual(
            to_json({"a": [1, 2, 3], 6: 'b'}), {"a": [1, 2, 3], "6": 'b'}
        )


class StrConstantTest(TestCase):

    def test_to_str_constant(self):
        self.assertEqual(to_str_constant('abc'), "'abc'")
        self.assertEqual(to_str_constant('abc', "'"), "'abc'")
        self.assertEqual(to_str_constant('abc', '"'), '"abc"')
        self.assertEqual(to_str_constant("a'bc", "'"), "'a\\'bc'")
        self.assertEqual(to_str_constant("a'bc", '"'), '"a\'bc"')
        self.assertEqual(to_str_constant('a"bc', '"'), '"a\\"bc"')
        self.assertEqual(to_str_constant('a"bc', "'"), "'a\"bc'")

    def test_to_str_constant_with_eval(self):
        s1 = '\\\''
        s2 = to_str_constant(s1, "'")
        self.assertEqual(s2, "'\\\\\\''")
        self.assertEqual(eval(s2), s1)
        s2 = to_str_constant(s1, '"')
        self.assertEqual(s2, '"\\\\\'"')
        self.assertEqual(eval(s2), s1)

    def test_is_str_constant(self):
        self.assertEqual(is_str_constant('abc'), False)
        self.assertEqual(is_str_constant('\\\''), False)
        self.assertEqual(is_str_constant('"abc"'), True)
        self.assertEqual(is_str_constant('"abc\''), False)
        self.assertEqual(is_str_constant("'abc'"), True)
        self.assertEqual(is_str_constant("\"abc'"), False)


class ToScalarTest(TestCase):

    def test_primitives(self):
        self.assertEqual(to_scalar(3456), 3456)
        self.assertEqual(to_scalar(34.56789), 34.56789)
        self.assertEqual(to_scalar(34.56789, num_digits=1), 34.6)
        self.assertEqual(to_scalar(True), True)
        self.assertEqual(to_scalar("Oh!"), "Oh!")
        self.assertEqual(to_scalar("Oh!Oh!", num_chars=2), "Oh...")
        self.assertEqual(to_scalar(None), None)

    def test_list(self):
        self.assertIs(to_scalar([]), UNDEFINED)
        self.assertEqual(to_scalar([1]), UNDEFINED)
        self.assertIs(to_scalar([1, 2, 3]), UNDEFINED)
        self.assertEqual(to_scalar(["Oh!"]), UNDEFINED)
        self.assertEqual(to_scalar([1, 2, 3], stringify=True), '[1, 2, 3]')

    def test_dict(self):
        self.assertIs(to_scalar({}), UNDEFINED)
        self.assertIs(to_scalar({'a': 1}), UNDEFINED)
        self.assertIs(to_scalar({'a': 1, 'b': 2}), UNDEFINED)
        self.assertEqual(
            to_scalar({'a': 1, 'b': 2}, stringify=True), "{'a': 1, 'b': 2}"
        )

    def test_ndarrays(self):
        self.assertIs(to_scalar(np.array([])), UNDEFINED)
        self.assertEqual(to_scalar(np.array(234)), 234)
        self.assertEqual(to_scalar(np.array([234])), 234)
        self.assertEqual(to_scalar(np.array([[234]])), 234)
        self.assertIs(to_scalar(np.array([234, 567])), UNDEFINED)
        self.assertIs(to_scalar(np.array([[234], [567]])), UNDEFINED)
        self.assertEqual(to_scalar(np.array([234.567])), 234.567)
        self.assertEqual(to_scalar(np.array(234.567)), 234.567)
        self.assertEqual(to_scalar(np.array([[234.567]])), 234.567)
        self.assertEqual(to_scalar(np.array([234.567, 567.234])), UNDEFINED)
        self.assertEqual(to_scalar(np.array([234.567]), num_digits=2), 234.57)
        self.assertEqual(to_scalar(np.array(True)), True)
        self.assertEqual(to_scalar(np.array([True])), True)
        self.assertIs(to_scalar(np.array([True, False])), UNDEFINED)
        self.assertIs(to_scalar(np.array([[True], [False]])), UNDEFINED)
        self.assertIs(to_scalar(np.array([None])), UNDEFINED)

    def test_xarrays(self):
        try:
            import xarray as xr
            self.assertIs(to_scalar(xr.DataArray(np.array([]))), UNDEFINED)
            self.assertEqual(to_scalar(xr.DataArray(np.array(234))), 234)
            self.assertEqual(to_scalar(xr.DataArray(np.array(234))), 234)
            self.assertEqual(to_scalar(xr.DataArray(np.array([234]))), 234)
            self.assertEqual(to_scalar(xr.DataArray(np.array([[234]]))), 234)
            self.assertIs(
                to_scalar(xr.DataArray(np.array([234, 567]))), UNDEFINED
            )
            self.assertIs(
                to_scalar(xr.DataArray(np.array([[234], [567]]))), UNDEFINED
            )
            self.assertEqual(
                to_scalar(xr.DataArray(np.array(234.567))), 234.567
            )
            self.assertEqual(
                to_scalar(xr.DataArray(np.array([234.567]))), 234.567
            )
            self.assertEqual(
                to_scalar(xr.DataArray(np.array([[234.567]]))), 234.567
            )
            self.assertEqual(
                to_scalar(xr.DataArray(np.array([234.567, 567.234]))),
                UNDEFINED
            )
            self.assertEqual(
                to_scalar(xr.DataArray(np.array([234.567])), num_digits=2),
                234.57
            )
            self.assertEqual(to_scalar(xr.DataArray(np.array(True))), True)
            self.assertEqual(to_scalar(xr.DataArray(np.array([True]))), True)
            self.assertIs(
                to_scalar(xr.DataArray(np.array([True, False]))), UNDEFINED
            )
            self.assertIs(
                to_scalar(xr.DataArray(np.array([[True], [False]]))), UNDEFINED
            )
        except ImportError:
            pass

    def test_pandas(self):
        try:
            import pandas as pd
            self.assertIs(to_scalar(pd.Series(np.array([]))), UNDEFINED)
            self.assertEqual(to_scalar(pd.Series(np.array([234]))), 234)
            self.assertIs(to_scalar(pd.Series(np.array([234, 567]))), UNDEFINED)
            self.assertEqual(to_scalar(pd.Series(np.array([234.567]))), 234.567)
            self.assertEqual(
                to_scalar(pd.Series(np.array([234.567, 567.234]))), UNDEFINED
            )
            self.assertEqual(
                to_scalar(pd.Series(np.array([234.567])), num_digits=2), 234.57
            )
            self.assertEqual(to_scalar(pd.Series(np.array([True]))), True)
            self.assertIs(
                to_scalar(pd.Series(np.array([True, False]))), UNDEFINED
            )

            self.assertIs(to_scalar(pd.Series([])), UNDEFINED)
            self.assertEqual(to_scalar(pd.Series(234)), 234)
            self.assertEqual(to_scalar(pd.Series([234])), 234)
            self.assertIs(to_scalar(pd.Series([234, 567])), UNDEFINED)
            self.assertEqual(to_scalar(pd.Series(234.567)), 234.567)
            self.assertEqual(to_scalar(pd.Series([234.567])), 234.567)
            self.assertEqual(
                to_scalar(pd.Series([234.567, 567.234])), UNDEFINED
            )
            self.assertEqual(
                to_scalar(pd.Series([234.567]), num_digits=2), 234.57
            )
            self.assertEqual(to_scalar(pd.Series([True])), True)
            self.assertIs(to_scalar(pd.Series([True, False])), UNDEFINED)
        except ImportError:
            pass



# The MIT License (MIT)
# Copyright (c) 2024 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

API data visualization operations

Components
==========

## plot_map

Plots a geospatial data slice on a world map. General python prompt usage:

```python
import matplotlib.pyplot as plt
from esa_climate_toolbox.ops.plot import plot_map

plot_map(dset)
plt.show()
```

General jupyter notebook usage:
```python
import matplotlib.pyplot as plt
from esa_climate_toolbox.ops.plot import plot_map

%matplotlib inline
plot_map(dset)
```
If a file path is given, the plot is saved.
Supported formats: eps, jpeg, jpg, pdf, pgf, png, ps, raw, rgba, svg, svgz, tif, tiff

"""

import matplotlib
import matplotlib.colors

import matplotlib.pyplot as plt
from matplotlib.figure import Figure

import xarray as xr
import cartopy.crs as ccrs
import numpy as np
import json

from esa_climate_toolbox.core.op import op
from esa_climate_toolbox.core.op import op_input
from esa_climate_toolbox.core.types import DatasetLike
from esa_climate_toolbox.core.types import DictLike
from esa_climate_toolbox.core.types import DimName
from esa_climate_toolbox.core.types import PolygonLike
from esa_climate_toolbox.core.types import ValidationError
from esa_climate_toolbox.core.types import VarName
from esa_climate_toolbox.core.types import VarNamesLike

from esa_climate_toolbox.ops.plot_helpers import get_var_data
from esa_climate_toolbox.ops.plot_helpers import get_vars_data
from esa_climate_toolbox.ops.plot_helpers import in_notebook
from esa_climate_toolbox.ops.plot_helpers import handle_plot_polygon
from esa_climate_toolbox.util.monitor import Monitor

# see https://matplotlib.org/tutorials/intermediate/tight_layout_guide.html
matplotlib.rcParams.update({'figure.autolayout': True})

# # noinspection PyBroadException
# try:
#     matplotlib.use('Qt5Agg')
#     has_qt5agg = True
# except Exception:
#     has_qt5agg = False

PLOT_FILE_EXTENSIONS = ['eps', 'jpeg', 'jpg', 'pdf', 'pgf',
                        'png', 'ps', 'raw', 'rgba', 'svg',
                        'svgz', 'tif', 'tiff']
PLOT_FILE_FILTER = dict(name='Plot Outputs', extensions=PLOT_FILE_EXTENSIONS)


@op(tags=['plot'], res_pattern='plot_{index}')
@op_input('ds')
@op_input('var', value_set_source='ds', data_type=VarName)
@op_input('indexers', data_type=DictLike)
@op_input('region', data_type=PolygonLike)
@op_input('projection', value_set=[
    'PlateCarree', 'LambertCylindrical', 'Mercator', 'Miller', 'Mollweide',
    'Orthographic', 'Robinson', 'Sinusoidal', 'NorthPolarStereo', 'SouthPolarStereo']
          )
@op_input('central_lon', units='degrees', value_range=[-180, 180])
@op_input('title')
@op_input('properties', data_type=DictLike)
@op_input('file', file_open_mode='w', file_filters=[PLOT_FILE_FILTER])
def plot_map(ds: xr.Dataset,
             var: VarName.TYPE = None,
             indexers: DictLike.TYPE = None,
             region: PolygonLike.TYPE = None,
             projection: str = 'PlateCarree',
             central_lon: float = 0.0,
             title: str = None,
             contour_plot: bool = False,
             properties: DictLike.TYPE = None,
             file: str = None) -> object:
    """
    Create a geographic map plot for the variable given by dataset *ds*
    and variable name *var*.

    Plots the given variable from the given dataset on a map with coastal lines.
    In case no variable name is given, the first encountered variable in the
    dataset is plotted. In case no *time* is given, the first time slice
    is taken. It is also possible to set extents of the plot. If no extents
    are given, a global plot is created.

    The plot can either be shown using pyplot functionality, or saved,
    if a path is given. The following file formats for saving the plot
    are supported: eps, jpeg, jpg, pdf, pgf, png, ps, raw, rgba, svg,
    svgz, tif, tiff

    :param ds: the dataset containing the variable to plot
    :param var: the variable's name
    :param indexers: Optional indexers into data array of *var*. The *indexers* is a
        dictionary or a comma-separated string of key-value pairs that maps the
        variable's dimension names to constant labels. e.g. "layer=4".
    :param region: Region to plot
    :param projection: name of a global projection,
        see http://scitools.org.uk/cartopy/docs/v0.15/crs/projections.html
    :param central_lon: central longitude of the projection in degrees
    :param title: an optional title
    :param contour_plot: If true plot a filled contour plot of data,
        otherwise plots a pixelated colormesh
    :param properties: optional plot properties for Python matplotlib,
           e.g. "bins=512, range=(-1.5, +1.5)"
           For full reference refer to
           https://matplotlib.org/api/lines_api.html and
           https://matplotlib.org/api/_as_gen/matplotlib.axes.Axes.contourf.html
    :param file: path to a file in which to save the plot
    :return: a matplotlib figure object or None if in IPython mode
    """
    if not isinstance(ds, xr.Dataset):
        raise ValidationError('Only gridded datasets are currently supported.')

    var_name = None
    if not var:
        for key in ds.data_vars.keys():
            var_name = key
            break
    else:
        var_name = VarName.convert(var)
    var = ds[var_name]

    indexers = DictLike.convert(indexers) or {}
    properties = DictLike.convert(properties) or {}

    extents = None
    bounds = handle_plot_polygon(region)
    if bounds:
        lon_min, lat_min, lon_max, lat_max = bounds
        extents = [lon_min, lon_max, lat_min, lat_max]

    if len(ds.lat) < 2 or len(ds.lon) < 2:
        # Matplotlib can not plot datasets with less than these dimensions with
        # contourf and pcolormesh methods
        raise ValidationError('The minimum dataset spatial dimensions to create a map'
                              ' plot are (2,2)')

    # See http://scitools.org.uk/cartopy/docs/v0.15/crs/projections.html#
    if projection == 'PlateCarree':
        proj = ccrs.PlateCarree(central_longitude=central_lon)
    elif projection == 'LambertCylindrical':
        proj = ccrs.LambertCylindrical(central_longitude=central_lon)
    elif projection == 'Mercator':
        proj = ccrs.Mercator(central_longitude=central_lon)
    elif projection == 'Miller':
        proj = ccrs.Miller(central_longitude=central_lon)
    elif projection == 'Mollweide':
        proj = ccrs.Mollweide(central_longitude=central_lon)
    elif projection == 'Orthographic':
        proj = ccrs.Orthographic(central_longitude=central_lon)
    elif projection == 'Robinson':
        proj = ccrs.Robinson(central_longitude=central_lon)
    elif projection == 'Sinusoidal':
        proj = ccrs.Sinusoidal(central_longitude=central_lon)
    elif projection == 'NorthPolarStereo':
        proj = ccrs.NorthPolarStereo(central_longitude=central_lon)
    elif projection == 'SouthPolarStereo':
        proj = ccrs.SouthPolarStereo(central_longitude=central_lon)
    else:
        raise ValidationError('illegal projection: "%s"' % projection)

    figure = plt.figure(figsize=(8, 4))
    ax = plt.axes(projection=proj)
    if extents:
        ax.set_extent(extents, ccrs.PlateCarree())
    else:
        ax.set_global()

    ax.coastlines()
    var_data = get_var_data(var, indexers, remaining_dims=('lon', 'lat'))

    # transform keyword is for the coordinate our data is in, which in case of a
    # 'normal' lat/lon dataset is PlateCarree.
    if contour_plot:
        var_data.plot.contourf(ax=ax, transform=ccrs.PlateCarree(), **properties)
    else:
        var_data.plot.pcolormesh(ax=ax, transform=ccrs.PlateCarree(), **properties)

    if title:
        ax.set_title(title)

    figure.tight_layout()

    if file:
        try:
            figure.savefig(file)
        except MemoryError:
            raise MemoryError(
                'Not enough memory to save the plot. Try using a different file format '
                'or enabling contour_plot.'
            )

    return figure if not in_notebook() else ax


@op(tags=['plot'], res_pattern='plot_{index}')
@op_input('var', value_set_source='ds', data_type=VarName)
@op_input('indexers', data_type=DictLike)
@op_input('title')
@op_input('filled')
@op_input('properties', data_type=DictLike)
@op_input('file', file_open_mode='w', file_filters=[PLOT_FILE_FILTER])
def plot_contour(ds: xr.Dataset,
                 var: VarName.TYPE,
                 indexers: DictLike.TYPE = None,
                 title: str = None,
                 filled: bool = True,
                 properties: DictLike.TYPE = None,
                 file: str = None) -> Figure:
    """
    Create a contour plot of a variable given by dataset *ds* and variable name *var*.

    :param ds: the dataset containing the variable to plot
    :param var: the variable's name
    :param indexers: Optional indexers into data array of *var*. The *indexers* is a
        dictionary or a comma-separated string of key-value pairs that maps the
        variable's dimension names to constant labels. e.g. "layer=4".
    :param title: an optional title
    :param filled: whether the regions between two contours shall be filled
    :param properties: optional plot properties for Python matplotlib,
           e.g. "bins=512, range=(-1.5, +1.5), label='Sea Surface Temperature'"
           For full reference refer to
           https://matplotlib.org/api/lines_api.html and
           https://matplotlib.org/devdocs/api/_as_gen/matplotlib.patches.Patch.html#matplotlib.patches.Patch
    :param file: path to a file in which to save the plot
    :return: a matplotlib figure object or None if in IPython mode
    """
    var_name = VarName.convert(var)
    if not var_name:
        raise ValidationError("Missing name for 'var'")
    var = ds[var_name]

    indexers = DictLike.convert(indexers) or {}
    properties = DictLike.convert(properties) or {}

    figure = plt.figure(figsize=(8, 4))
    ax = figure.add_subplot(111)

    var_data = get_var_data(var, indexers)
    if filled:
        var_data.plot.contourf(ax=ax, **properties)
    else:
        var_data.plot.contour(ax=ax, **properties)

    if title:
        ax.set_title(title)

    figure.tight_layout()

    if file:
        figure.savefig(file)

    return figure if not in_notebook() else None


@op(tags=['plot'], res_pattern='plot_{index}')
@op_input('ds', data_type=DatasetLike)
@op_input('var', value_set_source='ds', data_type=VarName)
@op_input('indexers', data_type=DictLike)
@op_input('title')
@op_input('properties', data_type=DictLike)
@op_input('file', file_open_mode='w', file_filters=[PLOT_FILE_FILTER])
def plot(ds: DatasetLike.TYPE,
         var: VarName.TYPE,
         indexers: DictLike.TYPE = None,
         title: str = None,
         properties: DictLike.TYPE = None,
         file: str = None) -> Figure:
    """
    Create a 1D/line or 2D/image plot of a variable given by dataset *ds*
    and variable name *var*.

    :param ds: Dataset or Dataframe that contains the variable named by *var*.
    :param var: The name of the variable to plot
    :param indexers: Optional indexers into data array of *var*. The *indexers* is a
        dictionary or a comma-separated string of key-value pairs that maps the
        variable's dimension names to constant labels,
        e.g., "lat=12.4, time='2012-05-02'".
    :param title: an optional plot title
    :param properties: optional plot properties for Python matplotlib,
        e.g., "bins=512, range=(-1.5, +1.5), label='Sea Surface Temperature'"
        For full reference refer to https://matplotlib.org/api/lines_api.html and
           https://matplotlib.org/devdocs/api/_as_gen/matplotlib.patches.Patch.html#matplotlib.patches.Patch
    :param file: path to a file in which to save the plot
    :return: a matplotlib figure object or None if in IPython mode
    """
    ds = DatasetLike.convert(ds)

    var_name = VarName.convert(var)
    if not var_name:
        raise ValidationError("Missing name for 'var'")
    var = ds[var_name]

    indexers = DictLike.convert(indexers)
    properties = DictLike.convert(properties) or {}

    figure = plt.figure()
    ax = figure.add_subplot(111)

    var_data = get_var_data(var, indexers)
    var_data.plot(ax=ax, **properties)

    if title:
        ax.set_title(title)

    figure.tight_layout()

    if file:
        figure.savefig(file)

    return figure if not in_notebook() else None


@op(tags=['plot'], res_pattern='plot_{index}')
@op_input('ds', data_type=DatasetLike)
@op_input('var_names', value_set_source='ds', data_type=VarNamesLike)
@op_input('fmt')
@op_input('label', value_set_source='ds.var_names', data_type=DimName)
@op_input('indexers', data_type=DictLike)
@op_input('title')
@op_input('file', file_open_mode='w', file_filters=[PLOT_FILE_FILTER])
def plot_line(ds: DatasetLike.TYPE,
              var_names: VarNamesLike.TYPE,
              fmt: str = None,
              label: DimName.TYPE = None,
              indexers: DictLike.TYPE = None,
              title: str = None,
              file: str = None) -> Figure:
    """
    Create a 1D/line plot of variable(s) given by dataset *ds*
    and variable name(s) *var_names*.

    :param ds: Dataset or Dataframe that contains the variable(s) named by *var_names*.
    :param var_names: The name of the variable(s) to plot
    :param fmt: optional semicolon-separated matplotlib formats, e.g.,
        1 variable - "b.-"
        2 variables - "b.-;r+:"
        If the number of properties is less than the number of selected variables,
        the next non-corresponding variable will repeat the first style on the list,
        and so on. For full reference on matplotlib plot() function, refer to
        https://matplotlib.org/api/_as_gen/matplotlib.pyplot.plot.html
    :param file: path to a file in which to save the plot
    :param label: dimension name to be selected as the x-axis of the plot
    :param indexers: Optional indexers into data array of *var_names*. The *indexers*
        is a dictionary or a comma-separated string of key-value pairs that maps the
        variable's dimension names to constant labels,
        e.g., "lat=12.4, time='2012-05-02'".
    :param title: an optional plot title
    :return: a matplotlib figure object or None if in IPython mode
    """
    ds = DatasetLike.convert(ds)

    fmt_count = 0
    fmt_list = []

    if fmt:
        fmt_list = fmt.split(";")
        fmt_count = len(fmt_list)

    if not var_names:
        raise ValidationError("Missing name for 'vars'")

    figure = plt.figure()
    ax = figure.add_subplot(111)
    figure.subplots_adjust(right=0.65)

    var_names = VarNamesLike.convert(var_names)
    if not title:
        if label:
            title = ','.join(var_names) + ' over ' + label
        else:
            title = ','.join(var_names)
    if indexers:
        title = title + '\n' + ' at ' + json.dumps(indexers).strip('"')
    ax.set_title(title)

    indexers = DictLike.convert(indexers)

    ax_var = {}
    var_count = len(var_names)
    predefined_fmt = ['r', 'g', 'b', 'c', 'm', 'y', 'k']
    if label:
        ds = get_vars_data(ds, indexers, remaining_dims=[label])
    else:
        ds = get_vars_data(ds, indexers)

    for i in range(var_count):
        var_name = var_names[i]
        var = ds[var_name]
        if len(var.dims) > 1:
            raise ValidationError(
                f'Unable to plot because variable {var_name} has more '
                f'than one dimension: {var.dims}. '
                f'To specify value(s) of these dimension(s), please use the indexers.'
            )

        var_label = var_name + ' (' + var.attrs['units'] + ')' \
            if 'units' in var.attrs else var_name
        properties_dict = {}

        indexers = DictLike.convert(indexers)

        if fmt is None:
            selected_fmt = predefined_fmt[i % len(predefined_fmt)]
        else:
            selected_fmt = fmt_list[i % fmt_count]

        if label:
            x_axis = var[label]
        elif 'time' in var:
            x_axis = var.time
        else:
            x_axis = []
        # to differentiate the creation of y-axis of the first and the nth variable
        if i == 0:
            if len(x_axis) > 0:
                ax.plot(x_axis, var, selected_fmt, **properties_dict)
            else:
                ax.plot(var, selected_fmt, **properties_dict)
            ax.set_ylabel(var_label, wrap=True)
            ax.yaxis.label.set_color(selected_fmt[0])
            ax.tick_params(axis='y', colors=selected_fmt[0])
        else:
            ax_var[var_name] = ax.twinx()
            if len(ax_var) > 1:
                ax_var[var_name].spines["right"].set_position(
                    ("axes", 1 + ((i - 1) * 0.2))
                )
                ax_var[var_name].set_frame_on(True)
                ax_var[var_name].patch.set_visible(False)
            if len(x_axis) > 0:
                ax_var[var_name].plot(x_axis, var, selected_fmt, **properties_dict)
            else:
                ax_var[var_name].plot(var, selected_fmt, **properties_dict)
            ax_var[var_name].set_ylabel(var_label, wrap=True)
            ax_var[var_name].yaxis.label.set_color(selected_fmt[0])
            ax_var[var_name].tick_params(axis='y', colors=selected_fmt[0])

    ax.tick_params(axis='x', rotation=45)
    if label in ds and 'long_name' in ds[label].attrs:
        ax.set_xlabel(ds[label].attrs['long_name'])
    figure.tight_layout()

    if file:
        figure.savefig(file, dpi=600)

    return figure if not in_notebook() else None


SCATTER_PLOT_TYPES = ['Point', 'Hexbin', '2D Histogram']


# noinspection PyShadowingBuiltins
@op(tags=['plot'], res_pattern='plot_{index}')
@op_input('ds1')
@op_input('ds2')
@op_input('var1', value_set_source='ds1', data_type=VarName)
@op_input('var2', value_set_source='ds2', data_type=VarName)
@op_input('indexers1', data_type=DictLike)
@op_input('indexers2', data_type=DictLike)
@op_input('type', value_set=SCATTER_PLOT_TYPES)
@op_input('title')
@op_input('properties', data_type=DictLike)
@op_input('file', file_open_mode='w', file_filters=[PLOT_FILE_FILTER])
def plot_scatter(ds1: xr.Dataset,
                 ds2: xr.Dataset,
                 var1: VarName.TYPE,
                 var2: VarName.TYPE,
                 indexers1: DictLike.TYPE = None,
                 indexers2: DictLike.TYPE = None,
                 type: str = '2D Histogram',
                 title: str = None,
                 properties: DictLike.TYPE = None,
                 file: str = None) -> Figure:
    """
    Create a scatter plot of two variables given by datasets *ds1*, *ds2*
    and the variable names *var1*, *var2*.

    :param ds1: Dataset that contains the variable named by *var1*.
    :param ds2: Dataset that contains the variable named by *var2*.
    :param var1: The name of the first variable to plot
    :param var2: The name of the second variable to plot
    :param indexers1: Optional indexers into data array *var1*. The *indexers1* is a
        dictionary or comma-separated string of key-value pairs that maps
        the variable's dimension names to constant labels,
        e.g., "lat=12.4, time='2012-05-02'".
    :param indexers2: Optional indexers into data array *var2*.
    :param type: The plot type.
    :param title: optional plot title
    :param properties: optional plot properties for Python matplotlib,
           e.g. "bins=512, range=(-1.5, +1.5), label='Sea Surface Temperature'"
           For full reference refer to
           https://matplotlib.org/api/lines_api.html and
           https://matplotlib.org/devdocs/api/_as_gen/matplotlib.patches.Patch.html#matplotlib.patches.Patch
    :param file: path to a file in which to save the plot
    :return: a matplotlib figure object or None if in IPython mode
    """
    var_name1 = VarName.convert(var1)
    var_name2 = VarName.convert(var2)
    indexers1 = DictLike.convert(indexers1) or {}
    indexers2 = DictLike.convert(indexers2) or {}
    properties = DictLike.convert(properties) or {}

    datasets = ds1, ds2
    var_names = var_name1, var_name2
    vars = [None, None]
    for i in (0, 1):
        try:
            vars[i] = datasets[i][var_names[i]]
        except KeyError as e:
            raise ValidationError(
                f'"{var_names[i]}" is not a variable in dataset given by "ds{i+1}"'
            ) from e

    var_dim_names = set(vars[0].dims), set(vars[1].dims)
    indexer_dim_names = set(indexers1.keys()), set(indexers2.keys())

    if set(var_dim_names[0]).isdisjoint(var_dim_names[1]):
        raise ValidationError('"var1" and "var2" have no dimensions in common:'
                              f' {var_dim_names[0]} and {var_dim_names[1]}.')

    for i in (0, 1):
        if indexer_dim_names[i] and not (indexer_dim_names[i] < var_dim_names[i]):
            raise ValidationError(
                f'"indexers{i+1}" must be a subset of the dimensions of "var{i+1}", '
                f'but {indexer_dim_names[i]} is not a subset of {var_dim_names[i]}.')

    rem_dim_names1 = var_dim_names[0] - indexer_dim_names[0]
    rem_dim_names2 = var_dim_names[1] - indexer_dim_names[1]
    if rem_dim_names1 != rem_dim_names2:
        raise ValidationError(
            'Remaining dimensions of data from "var1" must be equal to remaining '
            'dimensions of data from "var2", but {rem_dim_names1} is not equal to '
            '{rem_dim_names2}. You may need to use the indexers correctly.')

    indexers = indexers1, indexers2
    labels = [None, None]
    for i in (0, 1):
        # Note, long_name can be really long, too long.
        # name = vars[i].attrs.get('long_name', var_names[i])
        name = var_names[i]
        units = vars[i].attrs.get('units', '-')
        labels[i] = f'{name} ({units})'
        if indexers[i]:
            try:
                vars[i] = vars[i].sel(method='nearest', **indexers[i])
            except (KeyError, ValueError, TypeError) as e:
                raise ValidationError(
                    f'"indexers{i+1}" is not valid for "var{i+1}": {e}'
                ) from e
            labels[i] += " at " + ",".join(f"{key} = {value}"
                                           for key, value in indexers[i].items())

    shape1 = vars[0].shape
    shape2 = vars[1].shape
    if shape1 != shape2:
        raise ValidationError('Remaining shape of data from "var1" must be equal to'
                              ' remaining shape of data from "var2",'
                              f' but {shape1} is not equal to {shape2}.'
                              ' You may need to use the "coregister" operation first.')

    figure = plt.figure(figsize=(8, 8))
    ax = figure.add_subplot(111)

    try:
        x = vars[0].values.flatten()
        y = vars[1].values.flatten()
    except MemoryError as e:
        raise ValidationError(
            'Out of memory. Try using a data subset or specify indexers '
            'to reduce number of dimensions.'
        ) from e

    default_cmap = 'Reds'

    if type == 'Point':
        ax.grid(color='grey', linestyle='-', linewidth=0.25, alpha=0.5)
        if 'alpha' not in properties:
            properties['alpha'] = 0.25
        if 'markerfacecolor' not in properties:
            properties['markerfacecolor'] = '#880000'
        if 'markeredgewidth' not in properties:
            properties['markeredgewidth'] = 0.0
        if 'markersize' not in properties:
            properties['markersize'] = 5.0
        ax.plot(x, y, '.', **properties)
    elif type == '2D Histogram':
        if 'cmap' not in properties:
            properties['cmap'] = default_cmap
        if 'bins' not in properties:
            properties['bins'] = (256, 256)
        if 'norm' not in properties:
            properties['norm'] = matplotlib.colors.LogNorm()
        if 'range' not in properties:
            xrange = np.nanpercentile(x, [0, 100])
            yrange = np.nanpercentile(y, [0, 100])
            properties['range'] = [xrange, yrange]
        h, xedges, yedges, pc = ax.hist2d(x, y, **properties)
        figure.colorbar(pc, ax=ax, cmap=properties['cmap'])
    elif type == 'Hexbin':
        if 'cmap' not in properties:
            properties['cmap'] = default_cmap
        if 'gridsize' not in properties:
            properties['gridsize'] = (64, 64)
        if 'norm' not in properties:
            properties['norm'] = matplotlib.colors.LogNorm()
        x = np.ma.masked_invalid(x, copy=False)
        y = np.ma.masked_invalid(y, copy=False)
        collection = ax.hexbin(x, y, **properties)
        figure.colorbar(collection, ax=ax, cmap=properties['cmap'])

    ax.set_xlabel(labels[0])
    ax.set_ylabel(labels[1])

    if title:
        ax.set_title(title)

    # see https://matplotlib.org/tutorials/intermediate/tight_layout_guide.html
    figure.tight_layout()

    if file:
        figure.savefig(file)

    return figure if not in_notebook() else None


@op(tags=['plot'], res_pattern='plot_{index}')
@op_input('var', value_set_source='ds', data_type=VarName)
@op_input('indexers', data_type=DictLike)
@op_input('title')
@op_input('properties', data_type=DictLike)
@op_input('file', file_open_mode='w', file_filters=[PLOT_FILE_FILTER])
def plot_hist(ds: xr.Dataset,
              var: VarName.TYPE,
              indexers: DictLike.TYPE = None,
              title: str = None,
              properties: DictLike.TYPE = None,
              file: str = None) -> Figure:
    """
    Plot a variable, optionally save the figure in a file.

    The plot can either be shown using pyplot functionality, or saved,
    if a path is given. The following file formats for saving the plot
    are supported: eps, jpeg, jpg, pdf, pgf, png, ps, raw, rgba, svg, svgz, tif, tiff

    :param ds: Dataset that contains the variable named by *var*.
    :param var: The name of the variable to plot
    :param indexers: Optional indexers into data array of *var*. The *indexers* is a
        dictionary or a comma-separated string of key-value pairs that maps the
        variable's dimension names to constant labels,
        e.g., "lon=12.6, layer=3, time='2012-05-02'".
    :param title: an optional title
    :param properties: optional histogram plot properties for Python matplotlib,
        e.g., "bins=512, range=(-1.5, +1.5), label='Sea Surface Temperature'".
        For full reference refer to
        https://matplotlib.org/devdocs/api/_as_gen/matplotlib.pyplot.hist.html and
        https://matplotlib.org/devdocs/api/_as_gen/matplotlib.patches.Patch.html#matplotlib.patches.Patch
    :param file: path to a file in which to save the plot
    :return: a matplotlib figure object or None if in IPython mode
    """
    var_name = VarName.convert(var)
    if not var_name:
        raise ValidationError("Missing name for 'var'")

    var = ds[var]

    indexers = DictLike.convert(indexers)
    properties = DictLike.convert(properties) or {}

    figure = plt.figure(figsize=(8, 4))
    ax = figure.add_subplot(111)
    figure.tight_layout()

    var_data = get_var_data(var, indexers)
    var_data.plot.hist(ax=ax, **properties)

    if title:
        ax.set_title(title)

    figure.tight_layout()

    if file:
        figure.savefig(file)

    return figure if not in_notebook() else None



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from xcube.util.assertions import assert_given


def ect_init():
    # Plugin initializer.
    import yaml
    import os
    from xcube.core.store import DataStoreConfig
    from xcube.core.store import get_data_store_params_schema

    from esa_climate_toolbox.conf import get_data_stores_path
    from esa_climate_toolbox.conf.defaults import STORES_CONF_FILE
    from esa_climate_toolbox.core.common import default_user_agent
    from esa_climate_toolbox.core.ds import ECT_DATA_STORE_POOL

    dir_path = os.path.dirname(os.path.abspath(__file__))
    default_stores_file = os.path.join(dir_path, 'data/stores.yml')

    configured_store_configs = {}
    if os.path.exists(STORES_CONF_FILE):
        with open(STORES_CONF_FILE, 'r') as fp:
            configured_store_configs = yaml.safe_load(fp)
    with open(default_stores_file, 'r') as fp:
        default_store_configs = yaml.safe_load(fp)
    store_configs = configured_store_configs | default_store_configs

    for store_name, store_config in store_configs.items():
        store_id = store_config.get('store_id')
        assert_given(store_id, name='store_id', exception_type=RuntimeError)

        if store_id == 'file' \
                and 'store_params' in store_config \
                and store_config.get('store_params', {}).get('root') is None:
            root = os.environ.get('ECT_LOCAL_DATA_STORE_PATH',
                                  os.path.join(get_data_stores_path(),
                                               store_name))
            # Note: even if the root directory doesn't exist yet,
            # the xcube "file" data store will create it for us.
            store_config['store_params']['root'] = root

        store_params_schema = get_data_store_params_schema(store_id)
        if 'user_agent' in store_params_schema.properties:
            if 'store_params' not in store_config:
                store_config['store_params'] = {}
            store_config['store_params']['user_agent'] = default_user_agent()

        store_config = DataStoreConfig.from_dict(store_config)
        ECT_DATA_STORE_POOL.add_store_config(store_name, store_config)



import os.path
import sys
from unittest import TestCase

from esa_climate_toolbox.util.process import run_subprocess
from esa_climate_toolbox.util.process import ProcessOutputMonitor
from .test_monitor import RecordingMonitor

DIR = os.path.dirname(__file__)
MAKE_ENTROPY = os.path.join(DIR, '..', 'core', 'executables', 'mkentropy.py')


class ProcessTest(TestCase):
    def setUp(self):
        self.monitor = RecordingMonitor()
        self.stdout_lines = []
        self.stderr_lines = []
        self.line_count = 0
        self.started = False

    def on_started(self, process):
        self.assertEqual(process.returncode, None)
        self.started = True

    def store_stderr_line(self, line):
        self.stderr_lines.append(line.strip())

    def store_stdout_line(self, line):
        self.stdout_lines.append(line.strip())

    def store_stdout_line_and_cancel(self, line):
        if self.line_count == 3:
            self.monitor.cancel()
        self.store_stdout_line(line)
        self.line_count += 1

    def test_execute_with_handler(self):
        exit_code = run_subprocess([sys.executable, MAKE_ENTROPY, '5', '0.1'],
                                   stdout_handler=self.store_stdout_line,
                                   stderr_handler=self.store_stderr_line)
        self.assertEqual(exit_code, 0)
        self.assertEqual(self.stdout_lines,
                         ['mkentropy: Running 5 steps',
                          'mkentropy: Did 1 of 5 steps: 20.0%',
                          'mkentropy: Did 2 of 5 steps: 40.0%',
                          'mkentropy: Did 3 of 5 steps: 60.0%',
                          'mkentropy: Did 4 of 5 steps: 80.0%',
                          'mkentropy: Did 5 of 5 steps: 100.0%',
                          'mkentropy: Done making some entropy',
                          '']
                         )
        self.assertEqual(self.stderr_lines, [''])

    def test_execute_with_handler_failing(self):
        exit_code = run_subprocess(
            [sys.executable, MAKE_ENTROPY, '5', '0.1', '3'],
            stdout_handler=self.store_stdout_line,
            stderr_handler=self.store_stderr_line
        )
        self.assertEqual(exit_code, 1)
        self.assertEqual(self.stdout_lines,
                         ['mkentropy: Running 5 steps',
                          'mkentropy: Did 1 of 5 steps: 20.0%',
                          'mkentropy: Did 2 of 5 steps: 40.0%',
                          'mkentropy: Did 3 of 5 steps: 60.0%',
                          '']
                         )
        self.assertTrue(len(self.stderr_lines) > 2)
        self.assertEqual(
            self.stderr_lines[-2], 'RuntimeError: An intended error occurred!'
        )
        self.assertEqual(self.stderr_lines[-1], '')

    def test_execute_with_handler_and_cancellation(self):
        exit_code = run_subprocess(
            [sys.executable, MAKE_ENTROPY, '5', '0.2'],
            started_handler=self.on_started,
            stdout_handler=self.store_stdout_line_and_cancel,
            stderr_handler=self.store_stderr_line,
            is_cancelled=self.monitor.is_cancelled,
            cancelled_check_period=0.02,
            kill_on_cancel=False
        )
        self.assertTrue(self.started)
        self.assertTrue(exit_code != 0)
        self.assertTrue(len(self.stdout_lines) > 1)
        self.assertEqual(self.stderr_lines, [''])

    def test_execute_with_handler_and_kill_cancellation(self):
        exit_code = run_subprocess(
            [sys.executable, MAKE_ENTROPY, '5', '0.2'],
            started_handler=self.on_started,
            stdout_handler=self.store_stdout_line_and_cancel,
            stderr_handler=self.store_stderr_line,
            is_cancelled=self.monitor.is_cancelled,
            cancelled_check_period=0.02,
            kill_on_cancel=True
        )
        self.assertTrue(self.started)
        self.assertTrue(exit_code != 0)
        self.assertTrue(len(self.stdout_lines) > 1)
        self.assertTrue(len(self.stderr_lines) > 0)

    def test_execute_with_monitor(self):
        handler = ProcessOutputMonitor(
            self.monitor,
            started=r'mkentropy: Running (?P<total_work>\d+)',
            progress=r'mkentropy: Did (?P<work>\d+)',
            done=lambda line: 'Done' in line
        )
        exit_code = run_subprocess(
            [sys.executable, MAKE_ENTROPY, '5', '0.1'], stdout_handler=handler
        )
        self.assertEqual(exit_code, 0)
        self.assertEqual(self.monitor.records, [
            ('start', 'Process running', 5.0),
            ('progress', 1.0, None, 20),
            ('progress', 1.0, None, 40),
            ('progress', 1.0, None, 60),
            ('progress', 1.0, None, 80),
            ('progress', 1.0, None, 100),
            ('done',)
        ])



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from esa_climate_toolbox.conf import get_http_proxy

__author__ = "Chris Bernat (Telespazio VEGA UK Ltd), " \
             "Hans Permana (Brockmann Consult GmbH), " \
             "Norman Fomferra (Brockmann Consult GmbH)"


def initialize_proxy():
    """
    Initialize user defined proxy settings, read proxy setting from config file.
    Populates value to 3rd-party libraries using proper environment variables.
    """
    http_proxy = get_http_proxy()
    if not http_proxy:
        return

    from os import environ
    if http_proxy.startswith('https:'):
        environ['https_proxy'] = http_proxy
    elif http_proxy.startswith('http:'):
        environ['http_proxy'] = http_proxy


def configure_user_agent():
    """
    Sets proper Climate Toolbox HTTP User-Agent in 3rd-party HTTP libraries.
    Allows to mark HTTP requests sent from the ESA Climate Toolbox application
    via 3rd-party libraries as a part of application flow rather than
    showing them as an independent calls.
    """
    import requests.utils

    requests.utils.old_user_agent = requests.utils.default_user_agent
    requests.utils.default_user_agent = \
        lambda x=None: default_user_agent(requests.utils.old_user_agent(x)) \
            if x \
            else default_user_agent(requests.utils.old_user_agent())


def default_user_agent(ext: str = ""):
    """
    Default ESA Climate Toolbox HTTP User-Agent.
    :param ext:
    :return:
    """
    from ..version import __title__, __version__
    from platform import machine, python_version, system

    return "{title}/{version} (Python {python}; {system} {arch}) {ext}".format(
        title=__title__,
        version=__version__,
        python=python_version(),
        system=system(),
        arch=machine(),
        ext=ext)



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

Simple time-series extraction operation.

Functions
=========
"""

import xarray as xr

from esa_climate_toolbox.core.op import op
from esa_climate_toolbox.core.op import op_input
from esa_climate_toolbox.core.op import op_return
from esa_climate_toolbox.ops.select import select_var
from esa_climate_toolbox.core.types import PointLike
from esa_climate_toolbox.core.types import VarNamesLike
from esa_climate_toolbox.util.monitor import Monitor


@op(tags=['timeseries', 'temporal', 'filter', 'point'], version='1.0')
@op_input('point', data_type=PointLike)
@op_input('method', value_set=['nearest', 'ffill', 'bfill'])
@op_input('var', value_set_source='ds', data_type=VarNamesLike)
@op_return(add_history=True)
def tseries_point(ds: xr.Dataset,
                  point: PointLike.TYPE,
                  var: VarNamesLike.TYPE = None,
                  method: str = 'nearest') -> xr.Dataset:
    """
    Extract time-series from *ds* at given *lon*, *lat* position using
    interpolation *method* for each *var* given in a comma separated list of
    variables.

    The operation returns a new timeseries dataset, that contains the point
    timeseries for all required variables with original variable
    meta-information preserved.

    If a variable has more than three dimensions, the resulting timeseries
    variable will preserve all other dimensions except for lon/lat.

    :param ds: The dataset from which to perform timeseries extraction.
    :param point: Point to extract, e.g. (lon,lat)
    :param var: Variable(s) for which to perform the timeseries selection
                if none is given, all variables in the dataset will be used.
    :param method: Interpolation method to use.
    :return: A timeseries dataset
    """
    point = PointLike.convert(point)
    lon = point.x
    lat = point.y

    if not var:
        var = '*'

    retset = select_var(ds, var=var)
    indexers = {'lat': lat, 'lon': lon}
    retset = retset.sel(method=method, **indexers)

    # The dataset is no longer a spatial dataset -> drop associated global
    # attributes
    drop = ['geospatial_bounds_crs', 'geospatial_bounds_vertical_crs',
            'geospatial_vertical_min', 'geospatial_vertical_max',
            'geospatial_vertical_positive', 'geospatial_vertical_units',
            'geospatial_vertical_resolution', 'geospatial_lon_min',
            'geospatial_lat_min', 'geospatial_lon_max', 'geospatial_lat_max']

    for key in drop:
        retset.attrs.pop(key, None)

    return retset


@op(tags=['timeseries', 'temporal'], version='1.0')
@op_input('ds')
@op_input('var', value_set_source='ds', data_type=VarNamesLike)
@op_return(add_history=True)
def tseries_mean(ds: xr.Dataset,
                 var: VarNamesLike.TYPE,
                 mean_suffix: str = '_mean',
                 std_suffix: str = '_std',
                 monitor: Monitor = Monitor.NONE) -> xr.Dataset:
    """
    Extract spatial mean timeseries of the provided variables, return the
    dataset that in addition to all the information in the given dataset
    contains also timeseries data for the provided variables, following
    naming convention 'var_name1_ts_mean'. In addition, the standard deviation
    is computed.

    If a data variable with more dimensions than time/lat/lon is provided,
    the data will be reduced by taking the mean of all data values at a single
    time position resulting in one dimensional timeseries data variable.

    :param ds: The dataset from which to perform timeseries extraction.
    :param var: Variables for which to perform timeseries extraction
    :param mean_suffix: Mean suffix to use for resulting datasets
    :param std_suffix: Std suffix to use for resulting datasets
    :param monitor: a progress monitor.
    :return: Dataset with timeseries variables
    """
    if not var:
        var = '*'

    retset = select_var(ds, var)
    names = list(retset.data_vars.keys())

    with monitor.starting("Calculate mean and standard deviation",
                          total_work=len(names)):
        for name in names:
            dims = list(ds[name].dims)
            dims.remove('time')
            mean_name = name + mean_suffix
            with monitor.child(1).observing("Calculate mean"):
                retset[mean_name] = retset[name].mean(dim=dims, keep_attrs=True)
            retset[mean_name].attrs['ESA_Climate_Toolbox_Description'] = \
                'Mean aggregated over {} at each point in time.'.format(dims)
            std_name = name + std_suffix
            retset[std_name] = ds[name].std(dim=dims)
            retset[std_name].attrs['ESA_Climate_Toolbox_Description'] = \
                'Accompanying std values for variable \'{}\''.format(name)
    retset = retset.drop_vars(names)

    return retset



"""
Tests for subsetting operations
"""
from datetime import datetime
from unittest import TestCase

import numpy as np
import xarray as xr
import pandas as pd

from esa_climate_toolbox.core.op import OP_REGISTRY
from esa_climate_toolbox.core.opimpl import subset_spatial_impl
from esa_climate_toolbox.core.types import ValidationError
from esa_climate_toolbox.ops import subset
from esa_climate_toolbox.util.misc import object_to_qualified_name


def assert_dataset_equal(expected, actual):
    # this method is functionally equivalent to
    # `assert expected == actual`, but it checks each aspect
    # of equality separately for easier debugging
    assert expected.equals(actual), (expected, actual)


def get_test_subset_non_valid_lat_lon_dataset():
    temp = np.random.randn(2, 2, 3)
    precip = np.random.rand(2, 2, 3)
    lon = [[-40, 40], [-40, 40]]
    lat = [[-50, 50], [-50, 50]]
    return xr.Dataset({'temp': (['x', 'y', 'time'], temp),
                       'precip': (['x', 'y', 'time'], precip)},
                      coords={'lon': (['x', 'y'], lon),
                              'lat': (['x', 'y'], lat),
                              'time': pd.date_range('2014-09-06', periods=3)})


class TestSubsetSpatial(TestCase):

    def test_subset_non_valid_lat_lon(self):
        """
        Test whether lat and/or lon exist and if they exist whether
        they have dimension = 1
        :return: void
        """

        # Test whether lat lon exist
        dataset = xr.Dataset({
            'first': (['xc', 'yc', 'time'], np.ones([180, 360, 6])),
            'second': (['xc', 'yc', 'time'], np.ones([180, 360, 6])),
            'xc': np.linspace(-89.5, 89.5, 180),
            'yc': np.linspace(-179.5, 179.5, 360),
        })

        with self.assertRaises(ValidationError) as error:
            subset_spatial_impl(dataset, (-40, 40, -50, 50))
        self.assertIn('No (valid) geocoding found', str(error.exception))

        # test whether lat lon has the wrong dimension (!=1)
        dataset = get_test_subset_non_valid_lat_lon_dataset()

        with self.assertRaises(ValidationError) as error:
            subset_spatial_impl(dataset, (-40, 40, -50, 50))
        self.assertIn('Geocoding not recognised', str(error.exception))

    def test_nominal(self):
        """
        Test general 'most expected' use case functionality.
        """
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'lat': np.linspace(-89.5, 89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360)})
        actual = subset.subset_spatial(dataset, "-20, -10, 20, 10")
        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([22, 42, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([22, 42, 6])),
            'lat': np.linspace(-10.5, 10.5, 22),
            'lon': np.linspace(-20.5, 20.5, 42)})
        assert_dataset_equal(expected, actual)

    def test_inverted_dims_nominal(self):
        """
        Test if the implementation is dimension order agnostic.
        """
        # Inverted lat
        dataset = xr.Dataset({
            'first': (['lon', 'lat', 'time'], np.ones([360, 180, 6])),
            'second': (['lon', 'lat', 'time'], np.ones([360, 180, 6])),
            'lat': np.linspace(89.5, -89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360)})
        actual = subset.subset_spatial(dataset, "-20, -10, 20, 10")
        expected = xr.Dataset({
            'first': (['lon', 'lat', 'time'], np.ones([42, 22, 6])),
            'second': (['lon', 'lat', 'time'], np.ones([42, 22, 6])),
            'lat': np.linspace(10.5, -10.5, 22),
            'lon': np.linspace(-20.5, 20.5, 42)})
        assert_dataset_equal(expected, actual)

    def test_generic_masked(self):
        """
        Test using a generic Polygon and masking
        """
        # Africa
        a = str('POLYGON((-10.8984375 35.60371874069731,-19.16015625 '
                '23.885837699861995,-20.56640625 17.14079039331665,-18.6328125 '
                '7.536764322084079,-10.72265625 0.7031073524364783,10.37109375 '
                '0.3515602939922709,10.37109375 -22.268764039073965,22.8515625 '
                '-42.29356419217007,37.79296875 -27.21555620902968,49.39453125 '
                '-3.5134210456400323,54.4921875 14.093957177836236,18.984375 '
                '35.88905007936091,-10.8984375 35.60371874069731))')

        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'lat': np.linspace(-89.5, 89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360)})
        actual = subset.subset_spatial(dataset, a)
        # Gulf of Guinea
        gog = actual.sel(method='nearest', **{'lon': 1.2, 'lat': -1.4})
        self.assertTrue(np.isnan(gog['first']).all())
        # Africa
        self.assertTrue(1 == actual.sel(method='nearest', **{'lon': 20.7, 'lat': 6.15}))

    def test_generic_masked_inverted(self):
        """
        Test using a generic Polygon and masking
        """
        # Africa
        a = str('POLYGON((-10.8984375 35.60371874069731,-19.16015625 '
                '23.885837699861995,-20.56640625 17.14079039331665,-18.6328125 '
                '7.536764322084079,-10.72265625 0.7031073524364783,10.37109375 '
                '0.3515602939922709,10.37109375 -22.268764039073965,22.8515625 '
                '-42.29356419217007,37.79296875 -27.21555620902968,49.39453125 '
                '-3.5134210456400323,54.4921875 14.093957177836236,18.984375 '
                '35.88905007936091,-10.8984375 35.60371874069731))')

        # Inverted lat
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'lat': np.linspace(89.5, -89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360)})
        actual = subset.subset_spatial(dataset, a)
        # Gulf of Guinea
        gog = actual.sel(method='nearest', **{'lon': 1.2, 'lat': -1.4})
        self.assertTrue(np.isnan(gog['first']).all())
        # Africa
        self.assertTrue(1 == actual.sel(method='nearest', **{'lon': 20.7, 'lat': 6.15}))

    def test_generic_not_masked(self):
        """
        Test using a generic Polygon without masking
        """
        # Africa
        a = str('POLYGON((-10.8984375 35.60371874069731,-19.16015625 '
                '23.885837699861995,-20.56640625 17.14079039331665,-18.6328125 '
                '7.536764322084079,-10.72265625 0.7031073524364783,10.37109375 '
                '0.3515602939922709,10.37109375 -22.268764039073965,22.8515625 '
                '-42.29356419217007,37.79296875 -27.21555620902968,49.39453125 '
                '-3.5134210456400323,54.4921875 14.093957177836236,18.984375 '
                '35.88905007936091,-10.8984375 35.60371874069731))')

        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'lat': np.linspace(-89.5, 89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360)})
        actual = subset.subset_spatial(dataset, a, mask=False)
        # Gulf of Guinea
        self.assertTrue(1 == actual.sel(method='nearest', **{'lon': 1.2, 'lat': -1.4}))
        # Africa
        self.assertTrue(1 == actual.sel(method='nearest', **{'lon': 20.7, 'lat': 6.15}))

    def test_generic_not_masked_inverted(self):
        """
        Test using a generic Polygon without masking
        """
        # Africa
        a = str('POLYGON((-10.8984375 35.60371874069731,-19.16015625 '
                '23.885837699861995,-20.56640625 17.14079039331665,-18.6328125 '
                '7.536764322084079,-10.72265625 0.7031073524364783,10.37109375 '
                '0.3515602939922709,10.37109375 -22.268764039073965,22.8515625 '
                '-42.29356419217007,37.79296875 -27.21555620902968,49.39453125 '
                '-3.5134210456400323,54.4921875 14.093957177836236,18.984375 '
                '35.88905007936091,-10.8984375 35.60371874069731))')

        # Inverted lat
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'lat': np.linspace(89.5, -89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360)})
        actual = subset.subset_spatial(dataset, a, mask=False)
        # Gulf of Guinea
        self.assertTrue(1 == actual.sel(method='nearest', **{'lon': 1.2, 'lat': -1.4}))
        # Africa
        self.assertTrue(1 == actual.sel(method='nearest', **{'lon': 20.7, 'lat': 6.15}))

    def test_registered(self):
        """
        Test if it runs as an operation registered in the op registry.
        """
        reg_op = OP_REGISTRY.get_op(object_to_qualified_name(subset.subset_spatial))
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'lat': np.linspace(-89.5, 89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360)})
        actual = reg_op(ds=dataset, region="-20, -10, 20, 10")
        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([22, 42, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([22, 42, 6])),
            'lat': np.linspace(-10.5, 10.5, 22),
            'lon': np.linspace(-20.5, 20.5, 42)})
        assert_dataset_equal(expected, actual)

    def test_antimeridian_simple(self):
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'lat': np.linspace(-89.5, 89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360)})

        # With masking
        actual = subset.subset_spatial(dataset, '170, -5, -170, 5', mask=True)
        masked = actual.sel(method='nearest', **{'lon': 0, 'lat': 0})
        self.assertTrue(np.isnan(masked['first']).all())

    def test_antimeridian_simple_inverted(self):
        # Inverted lat
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'lat': np.linspace(89.5, -89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360)})

        # With masking
        actual = subset.subset_spatial(dataset, '170, -5, -170, 5', mask=True)
        masked = actual.sel(method='nearest', **{'lon': 0, 'lat': 0})
        self.assertTrue(np.isnan(masked['first']).all())

    def test_antimeridian_arbitrary(self):
        antimeridian_pol = str('POLYGON(('
                               '162.0703125 39.639537564366705,'
                               '-155.390625 39.774769485295465,'
                               '-155.56640625 12.726084296948184,'
                               '162.24609375 12.897489183755905,'
                               '161.89453125 26.745610382199025,'
                               '162.0703125 39.639537564366705'
                               '))')
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'lat': np.linspace(-89.5, 89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360)})

        with self.assertRaises(Exception) as cm:
            subset.subset_spatial(dataset, antimeridian_pol)
        self.assertIn('anti-meridian', str(cm.exception))

    def test_antimeridian_arbitrary_inverted(self):
        antimeridian_pol = str('POLYGON(('
                               '162.0703125 39.639537564366705,'
                               '-155.390625 39.774769485295465,'
                               '-155.56640625 12.726084296948184,'
                               '162.24609375 12.897489183755905,'
                               '161.89453125 26.745610382199025,'
                               '162.0703125 39.639537564366705'
                               '))')
        # Inverted lat
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'lat': np.linspace(89.5, -89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360)})

        with self.assertRaises(Exception) as cm:
            subset.subset_spatial(dataset, antimeridian_pol)
        self.assertIn('anti-meridian', str(cm.exception))

    def test_select_single_center(self):
        """
        Test subset spatial with a polygon that completely fits
        inside pixel bounds
        """
        # Masked
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([6, 12, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([6, 12, 3])),
            'lat': np.linspace(-75, 75, 6),
            'lon': np.linspace(-165, 165, 12),
            'time': [datetime(2000, x, 1) for x in range(1, 4)]})
        poly = ((32, 2), (34, 28), (58, 29), (54, 4))
        actual = subset.subset_spatial(dataset, region=poly)
        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([1, 1, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([1, 1, 3])),
            'lat': [15.],
            'lon': [45.],
            'time': [datetime(2000, x, 1) for x in range(1, 4)]})

        assert_dataset_equal(expected, actual)

        # Not masked
        actual = subset.subset_spatial(dataset, region=poly, mask=False)
        assert_dataset_equal(expected, actual)

    def test_select_single_vertice(self):
        """
        Test with a polygon that encloses a single pixel vertice,
        but no centers. That is, crosses four pixels
        """
        # Masked
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([6, 12, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([6, 12, 3])),
            'lat': np.linspace(-75, 75, 6),
            'lon': np.linspace(-165, 165, 12),
            'time': [datetime(2000, x, 1) for x in range(1, 4)]})
        poly = ((25, 2), (27, 40), (58, 38), (54, 4))
        actual = subset.subset_spatial(dataset, region=poly)
        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([2, 2, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([2, 2, 3])),
            'lat': [15., 45.],
            'lon': [15., 45.],
            'time': [datetime(2000, x, 1) for x in range(1, 4)]})

        assert_dataset_equal(expected, actual)

        # Not masked
        actual = subset.subset_spatial(dataset, region=poly, mask=False)
        assert_dataset_equal(expected, actual)

    def test_select_1d_lat(self):
        """
        Test with a polygon that runs over pixel boundaries in the
        latitude direction, resulting in selecting a single column
        """
        # Masked
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([6, 12, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([6, 12, 3])),
            'lat': np.linspace(-75, 75, 6),
            'lon': np.linspace(-165, 165, 12),
            'time': [datetime(2000, x, 1) for x in range(1, 4)]})
        poly = ((25, 2), (27, 47), (12, 50), (10, 4))
        actual = subset.subset_spatial(dataset, region=poly)
        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([2, 1, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([2, 1, 3])),
            'lat': [15., 45.],
            'lon': [15.],
            'time': [datetime(2000, x, 1) for x in range(1, 4)]})

        assert_dataset_equal(expected, actual)

        # Not masked
        actual = subset.subset_spatial(dataset, region=poly, mask=False)
        assert_dataset_equal(expected, actual)

    def test_select_1d_lon(self):
        """
        Test with a polygon that runs over pixel boundaries in the
        longitude direction. E.g., selects a single row
        """
        # Masked
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([6, 12, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([6, 12, 3])),
            'lat': np.linspace(-75, 75, 6),
            'lon': np.linspace(-165, 165, 12),
            'time': [datetime(2000, x, 1) for x in range(1, 4)]})
        poly = ((-28, 32), (-26, 58), (28, 52), (25, 33))
        actual = subset.subset_spatial(dataset, region=poly)
        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([1, 2, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([1, 2, 3])),
            'lat': [45.],
            'lon': [-15., 15.],
            'time': [datetime(2000, x, 1) for x in range(1, 4)]})

        assert_dataset_equal(expected, actual)

        # Not masked
        actual = subset.subset_spatial(dataset, region=poly, mask=False)
        assert_dataset_equal(expected, actual)

    def test_select_from_single_pixel(self):
        """
        Test subset spatial where the input dataset has only one pixel
        """

        # Masked
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([1, 1, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([1, 1, 3])),
            'lat': [45.],
            'lon': [15.],
            'time': [datetime(2000, x, 1) for x in range(1, 4)]})
        poly = ((-28, 32), (-26, 58), (28, 52), (25, 33))
        actual = subset.subset_spatial(dataset, region=poly)

        assert_dataset_equal(dataset, actual)

        # Not masked
        actual = subset.subset_spatial(dataset, region=poly, mask=False)
        assert_dataset_equal(dataset, actual)

    def test_out_of_bounds(self):
        """
        Test that an appropriate error is raised when the polygon
        wouldn't select any data from the original dataset
        """
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([1, 1, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([1, 1, 3])),
            'lat': [45.],
            'lon': [15.],
            'time': [datetime(2000, x, 1) for x in range(1, 4)]})
        poly = ((32, 32), (34, 58), (56, 56), (58, 33))

        with self.assertRaises(ValueError) as err:
            subset.subset_spatial(dataset, region=poly)
        self.assertIn('Can not select', str(err.exception))

        # Not masked
        with self.assertRaises(ValueError) as err:
            subset.subset_spatial(dataset, region=poly, mask=False)
        self.assertIn('Can not select', str(err.exception))

    def test_non_geospatial_variable(self):
        """
        Test that subsetting a dataset that contains a non-geospatial
        dataset skips such a dataset
        """
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'third': (['time'], np.ones([6])),
            'lat': np.linspace(-89.5, 89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360),
            'time': np.array([0, 1, 2, 3, 4, 5])})
        actual = subset.subset_spatial(dataset, "-20, -10, 20, 10")
        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([22, 42, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([22, 42, 6])),
            'third': (['time'], np.ones([6])),
            'lat': np.linspace(-10.5, 10.5, 22),
            'lon': np.linspace(-20.5, 20.5, 42),
            'time': np.array([0, 1, 2, 3, 4, 5])})
        assert_dataset_equal(expected, actual)

        poly = "POLYGON ((6.609745636041873 45.391851854914776, " \
               "19.720614826531428 45.391851854914776, " \
               "19.720614826531428 37.06301149556095, " \
               "6.609745636041874 37.06301149556095, " \
               "6.609745636041873 45.391851854914776))"
        actual = subset.subset_spatial(dataset, poly, mask=True)
        xr.testing.assert_equal(expected.third, actual.third)


class TestSubsetTemporal(TestCase):

    def test_subset_temporal(self):
        # Test general functionality
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'lat': np.linspace(-89.5, 89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360),
            'time': [datetime(2000, x, 1) for x in range(1, 7)]})
        actual = subset.subset_temporal(dataset, '2000-01-10, 2000-04-01')
        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 3])),
            'lat': np.linspace(-89.5, 89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360),
            'time': [datetime(2000, x, 1) for x in range(2, 5)]})
        assert_dataset_equal(expected, actual)

    def test_invalid_dtype(self):
        # Test passing in a MJD dataset
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'lat': np.linspace(-89.5, 89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360),
            'time': [2451544.5,
                     2451575.5,
                     2451604.5,
                     2451635.5,
                     2451665.5,
                     2451696.5]})
        with self.assertRaises(ValueError) as err:
            subset.subset_temporal(dataset, '2000-01-10, 2000-04-01')
        self.assertIn('type datetime', str(err.exception))

    def test_registered(self):
        """
        Test if it runs as an operation registered in the op registry.
        """
        reg_op = OP_REGISTRY.get_op(object_to_qualified_name(subset.subset_temporal))
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'lat': np.linspace(-89.5, 89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360),
            'time': [datetime(2000, x, 1) for x in range(1, 7)]})
        actual = reg_op(ds=dataset, time_range='2000-01-10, 2000-04-01')
        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 3])),
            'lat': np.linspace(-89.5, 89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360),
            'time': [datetime(2000, x, 1) for x in range(2, 5)]})
        assert_dataset_equal(expected, actual)


class TestSubsetTemporalIndex(TestCase):

    def test_subset_temporal_index(self):
        # Test general functionality
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'lat': np.linspace(-89.5, 89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360),
            'time': ['2000-01-01',
                     '2000-02-01',
                     '2000-03-01',
                     '2000-04-01',
                     '2000-05-01',
                     '2000-06-01']})
        actual = subset.subset_temporal_index(dataset, 2, 4)
        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 3])),
            'lat': np.linspace(-89.5, 89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360),
            'time': ['2000-03-01', '2000-04-01', '2000-05-01']})
        assert_dataset_equal(expected, actual)

    def test_registered(self):
        """
        Test if it runs as an operation registered in the op registry.
        """
        reg_op = OP_REGISTRY.get_op(
            object_to_qualified_name(subset.subset_temporal_index)
        )
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 6])),
            'lat': np.linspace(-89.5, 89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360),
            'time': ['2000-01-01',
                     '2000-02-01',
                     '2000-03-01',
                     '2000-04-01',
                     '2000-05-01',
                     '2000-06-01']})
        actual = reg_op(ds=dataset, time_ind_min=2, time_ind_max=4)
        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([180, 360, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([180, 360, 3])),
            'lat': np.linspace(-89.5, 89.5, 180),
            'lon': np.linspace(-179.5, 179.5, 360),
            'time': ['2000-03-01', '2000-04-01', '2000-05-01']})
        assert_dataset_equal(expected, actual)


class TestExtractPoint(TestCase):
    @classmethod
    def setUpClass(cls):
        v1_data = np.arange(18).reshape((3, 3, 2))
        v2_data = np.arange(100, 118).reshape((3, 3, 2))
        v3_data = np.arange(36).reshape((3, 3, 2, 2))
        v4_data = np.arange(50, 59).reshape((3, 3))
        cls._ds = xr.Dataset(
            {
                'v1': (['lat', 'lon', 'd1'], v1_data),
                'v2': (['lat', 'lon', 'd2'], v2_data),
                'v3': (['lat', 'lon', 'd1', 'd2'], v3_data),
                'v4': (['lat', 'lon'], v4_data)
            },
            coords={
                'lon': np.array([12, 13, 14], np.dtype('float64')),
                'lat': np.array([22, 23, 24], np.dtype('float64')),
                'd1': [1, 2],
                'd2': [datetime(2000, 3, 1), datetime(2000, 4, 1)]
            }
        )

    def test_all_extra_dims(self):
        result = subset.extract_point(
            self._ds, (12.2, 23.2),
            indexers={'d1': 2, 'd2': '2000-03-01'}
        )
        self.assertEqual(
            {'lat': 23.0, 'lon': 12.0, 'v1': 7.0, 'v2': 106.0, 'v3': 14.0, 'v4': 53.0},
            result
        )

    def test_one_extra_dims(self):
        result = subset.extract_point(self._ds, (12.2, 23.2), indexers={'d1': 2})
        self.assertEqual({'lat': 23.0, 'lon': 12.0, 'v1': 7.0, 'v4': 53.0}, result)

        result = subset.extract_point(
            self._ds, (12.2, 23.2), indexers={'d2': '2000-03-01'}
        )
        self.assertEqual({'lat': 23.0, 'lon': 12.0, 'v2': 106.0, 'v4': 53.0}, result)

    def test_no_extra_dims(self):
        result = subset.extract_point(self._ds, (12.2, 23.2), indexers={})
        self.assertEqual({'lat': 23.0, 'lon': 12.0, 'v4': 53.0}, result)

        result = subset.extract_point(self._ds, (12.2, 23.2))
        self.assertEqual({'lat': 23.0, 'lon': 12.0, 'v4': 53.0}, result)

    def test_unknown_dim(self):
        result = subset.extract_point(self._ds, (12.2, 23.2), indexers={'x1': 42})
        self.assertEqual({'lat': 23.0, 'lon': 12.0, 'v4': 53.0}, result)

    def test_point_out_of_bounds(self):
        result = subset.extract_point(
            self._ds, (0, 0), indexers={'d1': 2, 'd2': '2000-03-01'}
        )
        self.assertEqual({}, result)

        result = subset.extract_point(self._ds, (0, 0))
        self.assertEqual({}, result)

    def test_extra_dim_with_no_exact_match(self):
        # no exact match for 'd2', the same as if 'd2' is not given
        result = subset.extract_point(
            self._ds, (12.2, 23.2), indexers={'d1': 2, 'd2': '2000-03-02'}
        )
        self.assertEqual({'lat': 23.0, 'lon': 12.0, 'v1': 7.0, 'v4': 53.0}, result)

        result = subset.extract_point(
            self._ds, (12.2, 23.2), indexers={'d1': 1.1, 'd2': '2000-03-01'}
        )
        self.assertEqual({'lat': 23.0, 'lon': 12.0, 'v2': 106.0, 'v4': 53.0}, result)



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__author__ = "Norman Fomferra (Brockmann Consult GmbH)"

import os.path

from ..version import __version__

DEFAULT_DATA_DIR_NAME = '.ect'
DEFAULT_DATA_PATH = os.path.join(os.path.expanduser('~'), DEFAULT_DATA_DIR_NAME)
DEFAULT_VERSION_DATA_PATH = os.path.join(DEFAULT_DATA_PATH, __version__)

GLOBAL_CONF_FILE = os.path.join(DEFAULT_DATA_PATH, 'conf.py')
STORES_CONF_FILE = os.environ.get(
    'ECT_STORES_CONFIG_PATH', os.path.join(DEFAULT_DATA_PATH, 'stores.yml')
)
VERSION_CONF_FILE = os.path.join(DEFAULT_VERSION_DATA_PATH, 'conf.py')
LOCAL_CONF_FILE = 'ect-conf.py'
LOCATION_FILE = 'ect.location'
USER_PREFERENCES_FILE = 'user-preferences.json'

DEFAULT_RES_PATTERN = 'res_{index}'

NETCDF_COMPRESSION_LEVEL = 9

_ONE_MIB = 1024 * 1024
_ONE_GIB = 1024 * _ONE_MIB

DEFAULT_COLOR_MAP = 'inferno'



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

This module defines the :py:class:`Monitor` interface that may be used by
functions and methods that offer support for observation and control of
long-running tasks.

The module also provides a simple but still useful default implementation
:py:class:`ConsoleMonitor`, which prints progress output directly to the
console.

Components
==========
"""

import signal
import sys
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from shutil import get_terminal_size

__author__ = "Norman Fomferra (Brockmann Consult GmbH)"

_DEBUG_DASK_PROGRESS = False
_DaskMonitor = None
_IS_DASK_AVAILABLE = None


class Cancellation(Exception):
    """The operations was cancelled by the user."""
    pass


class Monitor(metaclass=ABCMeta):
    """
    A monitor is used to both observe and control a running task.

    The ``Monitor`` class is an abstract base class for concrete monitors.
    Derived classes must implement the following three abstract methods:
    :py:meth:`start`, :py:meth:`progress`, and :py:meth:`done`.
    Derived classes must implement also the following two abstract methods,
    if they want cancellation support:
    :py:meth:`cancel` and :py:meth:`is_cancelled`.

    Pass ``Monitor.NONE`` to functions that expect a monitor instead of passing
    ``None``.

    Given here is an example of how progress monitors should be used by
    functions:::

        def long_running_task(a, b, c, monitor):
            with monitor.starting('doing a long running task', total_work=100)
                # do 30% of the work here
                monitor.progress(work=30)
                # do 70% of the work here
                monitor.progress(work=70)

    If a function makes calls to other functions that also support a monitor,
    a *child-monitor* is used:::

        def long_running_task(a, b, c, monitor):
            with monitor.starting('doing a long running task', total_work=100)
                # let other_task do 30% of the work
                other_task(a, b, c, monitor=monitor.child(work=30))
                # let other_task do 70% of the work
                other_task(a, b, c, monitor=monitor.child(work=70))

    """

    #: A valid monitor that effectively does nothing. Use ``Monitor.NONE``
    #: it instead of passing ``None`` to functions and methods that expect
    # an argument of type ``Monitor``.
    NONE = None

    @contextmanager
    def starting(self, label: str, total_work: float = None):
        """
        A context manager for easier use of progress monitors.
        Calls the monitor's ``start`` method with *label* and *total_work*.
        Will then take care of calling :py:meth:`Monitor.done`.

        :param label: Passed to the monitor's ``start`` method
        :param total_work: Passed to the monitor's ``start`` method
        :return:
        """
        self.start(label, total_work=total_work)
        try:
            yield self
        finally:
            self.done()

    @contextmanager
    def observing(self, label: str):
        """
        A context manager for easier use of progress monitors.
        Observes a ``dask`` task and reports back to the monitor.

        :param label: Passed to the monitor's ``start`` method
        :return:
        """
        dask_monitor = _get_dask_monitor()
        if dask_monitor is not None:
            with dask_monitor(label=label, monitor=self):
                yield
        else:
            raise NotImplementedError('Monitor.observing() requires "dask" '
                                      'package to be installed')

    @abstractmethod
    def start(self, label: str, total_work: float = None):
        """
        Call to signal that a task has started.

        Note that *label* and *total_work* are not passed to ``__init__``,
        because they are usually not known at constructions time. It is the
        responsibility of the task to derive the appropriate values for these.


        :param label: A task label
        :param total_work: The total amount of work
        """

    @abstractmethod
    def progress(self, work: float = None, msg: str = None):
        """
        Call to signal that a task has mode some progress.

        :param work: The incremental amount of work.
        :param msg: A detail message.
        """

    @abstractmethod
    def done(self):
        """
        Call to signal that a task has been done.
        """

    def child(self, work: float = 1) -> 'Monitor':
        """
        Return a child monitor for the given partial amount of *work*.

        :param work: The partial amount of work.
        :return: a sub-monitor
        """
        return ChildMonitor(self, work)

    def cancel(self):
        """
        Request the task to be cancelled.
        This method will be usually called from the code that created the
        monitor, not by users of the monitor. For example, a GUI could create
        the monitor due to an invocation of a long-running task, and then
        the user wishes to cancel that task. The default implementation does
        nothing.
        Override to implement something useful.
        """
        pass

    def is_cancelled(self) -> bool:
        """
        Check if there is an external request to cancel the current task
        observed by this monitor.

        Users of a monitor shall frequently call this method and check its
        return value. If cancellation is requested, they should politely exit
        the current processing in a proper way, e.g., by cleaning up allocated
        resources.
        The default implementation returns ``False``.
        Subclasses shall override this method to return ``True`` if a task
        cancellation request was detected.

        :return: ``True`` if task cancellation was requested externally.
            The default implementation returns ``False``.
        """
        return False

    def check_for_cancellation(self):
        """
        Checks if the monitor has been cancelled and raises a ``Cancellation``
        in that case.
        """
        if self.is_cancelled():
            raise Cancellation()


# noinspection PyAbstractClass
class _NullMonitor(Monitor):
    def __repr__(self):
        # Overridden to make Sphinx use a readable name.
        return 'Monitor.NONE'

    def start(self, label: str, total_work: float = None):
        pass

    def progress(self, work: float = None, msg: str = None):
        pass

    def done(self):
        pass

    def child(self, work: float = 1):
        return Monitor.NONE


#: Pass ``Monitor.NONE`` to functions that expect a monitor instead of passing
# ``None``.
Monitor.NONE = _NullMonitor()


# noinspection PyAbstractClass
class ChildMonitor(Monitor):
    """
    A child monitor is responsible for a partial amount of work of a
    *parent_monitor*.


    :param parent_monitor: the parent monitor
    :param partial_work: the partial amount of work of *parent_monitor*.
    """

    def __init__(self, parent_monitor: Monitor, partial_work: float):
        self._parent_monitor = parent_monitor
        self._partial_work = partial_work
        self._total_work = None
        self._label = None

    def start(self, label: str, total_work: float = None):
        self._label = label
        self._total_work = total_work
        parent_work = 0.0 if total_work is not None else None
        self._parent_monitor.progress(work=parent_work, msg=self._label)

    def progress(self, work: float = None, msg: str = None):
        parent_work = self._partial_work * (work / self._total_work) \
            if work is not None else None
        parent_msg = '%s: %s' % (self._label, msg) \
            if self._label and msg else None
        self._parent_monitor.progress(work=parent_work, msg=parent_msg)

    def done(self):
        parent_work = 0.0 if self._total_work is not None else None
        self._parent_monitor.progress(work=parent_work, msg=self._label)

    def cancel(self):
        self._parent_monitor.cancel()

    def is_cancelled(self) -> bool:
        return self._parent_monitor.is_cancelled()


# noinspection PyAbstractClass
class ConsoleMonitor(Monitor):
    """
    A simple console monitor that directly writes to ``sys.stdout`` and
    detects user cancellation requests via CTRL+C.

    :param stay_in_line: If ``True``,
        the text written out will stay in the same line.
    :param progress_bar_size: If ``> 1``, a progress monitor of
        max. *progress_bar_size* characters will be written to the console.
    """

    def __init__(self, stay_in_line=False, progress_bar_size=1):
        self._stay_in_line = stay_in_line
        self._progress_bar_size = progress_bar_size
        self._effective_progress_size = 0
        self._label = None
        self._worked = None
        self._total_work = None
        self._percentage = None
        self._cancelled = False
        self._last_line_len = None
        self._old_ctrl_c_handler = False
        self._msg = None
        self._term_size = 0

    def __del__(self):
        self._register_ctrl_c_handler(self._old_ctrl_c_handler)

    def start(self, label: str, total_work: float = None):
        self.check_for_cancellation()
        if not label:
            raise ValueError('label must be given')
        self._label = label
        self._worked = 0.
        self._percentage = None
        self._total_work = total_work
        self._register_ctrl_c_handler(self._on_ctrl_c)
        self._report_progress(msg='started')

    def progress(self, work: float = None, msg: str = None):
        self.check_for_cancellation()
        percentage = None
        if work is not None:
            self._worked += work
            percentage = self._calc_percentage()
        # only display progress on integer percentage change or message change
        if percentage != self._percentage or msg != self._msg:
            self._report_progress(percentage, msg)
        self._percentage = percentage
        self._msg = msg

    def done(self):
        self.check_for_cancellation()
        self._register_ctrl_c_handler(self._old_ctrl_c_handler)
        if self.is_cancelled():
            self._report_progress(msg='cancelled')
        else:
            self._report_progress(
                msg='done' if self._msg is None else self._msg
            )
        if self._stay_in_line:
            sys.stdout.write('\n')

    def cancel(self):
        self._cancelled = True

    def is_cancelled(self) -> bool:
        return self._cancelled

    def _report_progress(self, percentage=None, msg=None):
        term_size = get_terminal_size().columns
        if sys.platform == 'win32':
            term_size -= 1

        label_text = '%s: ' % self._label
        percentage_text = '' if percentage is None else '%3d%% ' % percentage
        progress_bar_text = ''
        message_text = 'in progress...' \
            if (not msg and percentage is None) else (msg or '')

        progress_bar_size = self._progress_bar_size
        min_text_size = \
            len(label_text) + len(percentage_text) + len(message_text)
        if percentage is not None and progress_bar_size > 1:
            available_progress_bar_size = term_size - (min_text_size + 3)
            if available_progress_bar_size > 1:
                if progress_bar_size > available_progress_bar_size:
                    progress_bar_size = available_progress_bar_size
            else:
                progress_bar_size = 0
            if progress_bar_size > 1:
                done_count = int(progress_bar_size * percentage / 100 + 0.5)
                remaining_count = progress_bar_size - done_count
                progress_bar_text = \
                    '[%s%s] ' % ('#' * done_count, '-' * remaining_count)

        line = label_text + progress_bar_text + percentage_text + message_text

        if self._stay_in_line:
            if len(line) > term_size:
                line = line[:term_size]
            else:
                line = line.ljust(term_size, ' ')
            sys.stdout.write('\r')
            sys.stdout.write(line)
        else:
            sys.stdout.write(line)
            sys.stdout.write('\n')
        sys.stdout.flush()

    def _calc_percentage(self):
        return round(100. * self._worked / self._total_work) \
            if self._worked is not None and self._total_work is not None \
            else None

    # noinspection PyUnusedLocal,PyShadowingNames
    def _on_ctrl_c(self, signal, frame):
        self.cancel()

    # noinspection PyMethodMayBeStatic
    def _register_ctrl_c_handler(self, ctrl_c_handler):
        if ctrl_c_handler:
            try:
                signal.signal(signal.SIGINT, ctrl_c_handler)
            except ValueError:
                # If not on main thread, we may receive ValueError:
                # signal only works in main thread
                pass


def _get_dask_monitor():
    global _DaskMonitor
    global _IS_DASK_AVAILABLE

    if _DaskMonitor is None and _IS_DASK_AVAILABLE is None:
        try:
            # noinspection PyUnresolvedReferences,PyPackageRequirements
            from dask.callbacks import Callback

            _IS_DASK_AVAILABLE = True

            class _DaskMonitor(Callback):
                """
                A ``dask.Callback`` that reports the task level notification
                that the dask scheduler generates to the provided ``Monitor``.

                This allows for tracking then progress inside dask compute/get
                calls and the possibility to cancel them.
                """

                def __init__(self, label: str, monitor: Monitor):
                    super().__init__()
                    self._label = label
                    self._monitor = monitor
                    self._is_done = False

                # noinspection PyUnusedLocal
                def _start_state(self, dsk, state):
                    if self._is_done:
                        return
                    num_tasks = sum(len(state[k]) for k in ['ready', 'waiting'])
                    self._monitor.start(label=self._label, total_work=num_tasks)
                    if _DEBUG_DASK_PROGRESS:
                        print("DaskMonitor.start_state: num_tasks=", num_tasks)

                # noinspection PyUnusedLocal
                def _posttask(self, key, result, dsk, state, worker_id):
                    if self._is_done:
                        return
                    self._monitor.progress(work=1)
                    if _DEBUG_DASK_PROGRESS:
                        print("DaskMonitor.posttask: key=", key)

                # noinspection PyUnusedLocal
                def _finish(self, dsk, state, failed):
                    if self._is_done:
                        return
                    self._monitor.done()
                    self._is_done = True
                    if _DEBUG_DASK_PROGRESS:
                        print("DaskMonitor.finish")

        except ImportError:
            _IS_DASK_AVAILABLE = False

    return _DaskMonitor



from unittest import TestCase
import os.path
from esa_climate_toolbox.util.tmpfile import del_temp_file
from esa_climate_toolbox.util.tmpfile import del_temp_files
from esa_climate_toolbox.util.tmpfile import get_temp_files
from esa_climate_toolbox.util.tmpfile import new_temp_file


class TempFileTest(TestCase):

    def setUp(self):
        del_temp_files(force=True)
        self.assertEqual(get_temp_files(), [])

    def test_all(self):

        p1 = new_temp_file(prefix='test-')
        p2 = new_temp_file(suffix='.nc')
        self.assertEqual(len(get_temp_files()), 2)
        self.assertEqual(get_temp_files()[0], p1)
        self.assertEqual(get_temp_files()[1], p2)
        self.assertTrue(
            os.path.basename(get_temp_files()[0][1]).startswith('test-')
        )
        self.assertTrue(
            os.path.basename(get_temp_files()[1][1]).endswith('.nc')
        )

        del_temp_files()
        self.assertEqual(len(get_temp_files()), 0)

        p1 = new_temp_file(prefix='test-')
        p2 = new_temp_file(suffix='.nc')
        del_temp_file(p1[1])
        self.assertEqual(len(get_temp_files()), 1)
        del_temp_file(p2[1])
        self.assertEqual(len(get_temp_files()), 0)



# The MIT License (MIT)
# Copyright (c) 2024 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

Arithmetic operations

Functions
=========
"""
import dask.array as da
import xarray as xr

from esa_climate_toolbox.core.op import op
from esa_climate_toolbox.core.op import op_input
from esa_climate_toolbox.core.op import op_return
from esa_climate_toolbox.core.types import DatasetLike
from esa_climate_toolbox.core.types import ValidationError
from esa_climate_toolbox.util.monitor import Monitor


@op(tags=['arithmetic'], version='1.0')
@op_input('ds', data_type=DatasetLike)
@op_return(add_history=True)
def arithmetics(
        ds: DatasetLike.TYPE, ops: str, monitor: Monitor = Monitor.NONE
) -> xr.Dataset:
    """
    Do arithmetic operations on the given dataset by providing a list of
    arithmetic operations and the corresponding constant. The operations will
    be applied to the dataset in the order in which they appear in the list.
    For example:
    'log,+5,-2,/3,*2'

    Currently supported arithmetic operations:
    log,log10,log2,log1p,exp,+,-,/,*

    where:
        log - natural logarithm
        log10 - base 10 logarithm
        log2 - base 2 logarithm
        log1p - log(1+x)
        exp - the exponential

    The operations will be applied element-wise to all arrays of the dataset.

    :param ds: The dataset to which to apply arithmetic operations
    :param ops: A comma separated list of arithmetic operations to apply
    :param monitor: a progress monitor.
    :return: The dataset with given arithmetic operations applied
    """
    ds = DatasetLike.convert(ds)
    retset = ds
    with monitor.starting('Calculate result', total_work=len(ops.split(','))):
        for item in ops.split(','):
            with monitor.child(1).observing("Calculate"):
                item = item.strip()
                if item[0] == '+':
                    retset = retset + float(item[1:])
                elif item[0] == '-':
                    retset = retset - float(item[1:])
                elif item[0] == '*':
                    retset = retset * float(item[1:])
                elif item[0] == '/':
                    retset = retset / float(item[1:])
                elif item[:] == 'log':
                    retset = da.log(retset)
                elif item[:] == 'log10':
                    retset = da.log10(retset)
                elif item[:] == 'log2':
                    retset = da.log2(retset)
                elif item[:] == 'log1p':
                    retset = da.log1p(retset)
                elif item[:] == 'exp':
                    retset = da.exp(retset)
                else:
                    raise ValidationError('Arithmetic operation {} not'
                                          ' implemented.'.format(item[0]))

    return retset


@op(tags=['arithmetic'], version='1.0')
@op_return(add_history=True)
def diff(ds: xr.Dataset,
         ds2: xr.Dataset,
         monitor: Monitor = Monitor.NONE) -> xr.Dataset:
    """
    Calculate the difference of two datasets (ds - ds2). This is done by
    matching variable names in the two datasets against each other and taking
    the difference of matching variables.

    If lat/lon/time extents differ between the datasets, the default behavior
    is to take the intersection of the datasets and run subtraction on that.
    However, broadcasting is possible. E.g. ds(lat/lon/time) - ds(lat/lon) is
    valid. In this case the subtrahend will be stretched to the size of
    ds(lat/lon/time) so that it can be subtracted. This also works if the
    subtrahend is a single time slice of arbitrary temporal position. In this
    case, the time dimension will be squeezed out leaving a lat/lon dataset.

    :param ds: The minuend dataset
    :param ds2: The subtrahend dataset
    :param monitor: a progress monitor.
    :return: The difference dataset
    """
    try:
        # Times do not intersect
        if 0 == len(ds.time - ds2.time) and \
                len(ds.time) == len(ds2.time):  # Times are the same length
            # If the datasets don't intersect in time dimension, a naive difference
            # would return empty data variables. Hence, the time coordinate has to
            # be dropped beforehand
            ds = ds.drop_vars('time')
            ds2 = ds2.drop_vars('time')
            return ds - ds2
    except AttributeError:
        # It is likely that the one operand is a lat/lon array that can be
        # broadcast against the other operand
        pass

    try:
        if 1 == len(ds2.time):
            # The subtrahend is a single time-slice -> squeeze 'time' dimension to
            # be able to broadcast is along minuend
            ds2 = ds2.squeeze('time', drop=True)
    except AttributeError:
        # Doesn't have a time dimension already
        pass
    except TypeError as e:
        if 'unsized object' in str(e):
            # The 'time' variable is a scalar
            pass
        else:
            raise TypeError(str(e))

    with monitor.observing("Subtract datasets"):
        ds_diff = ds - ds2

    return ds_diff



# The MIT License (MIT)
# Copyright (c) 2024 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import dask.array as da
import functools
import geopandas as gpd
import math
import pandas as pd
import pyproj
import shapely
import shapely.ops
import xarray as xr
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
import warnings

from xcube.core.extract import get_cube_values_for_points
from xcube.core.gridmapping import GridMapping

from esa_climate_toolbox.core.op import op
from esa_climate_toolbox.core.op import op_input
from esa_climate_toolbox.core.types import DataFrameLike
from esa_climate_toolbox.core.types import DatasetLike
from esa_climate_toolbox.core.types import GeometryLike
from esa_climate_toolbox.core.types import GeoDataFrame
from esa_climate_toolbox.core.types import Literal
from esa_climate_toolbox.core.types import PolygonLike
from esa_climate_toolbox.core.types import ValidationError
from esa_climate_toolbox.core.types import VarName
from esa_climate_toolbox.core.types import VarNamesLike
from esa_climate_toolbox.util.monitor import Monitor


_DEG2RAD = math.pi / 180.
_REGION_MODES = [
    'almost_equals',
    'contains',
    'crosses',
    'disjoint',
    'intersects',
    'touches',
    'within',
]

ReprojectionFunc = Callable[[float, float], Tuple[float, float]]


@op(tags=['statistics'], version='1.0')
@op_input('df', data_type=DataFrameLike)
@op_input('var_names', value_set_source='df', data_type=VarNamesLike)
@op_input('aggregate_geometry', data_type=bool)
def aggregate_statistics(df: DataFrameLike.TYPE,
                         var_names: VarNamesLike.TYPE = None,
                         aggregate_geometry: bool = False,
                         monitor: Monitor = Monitor.NONE) -> pd.DataFrame:
    """
    Aggregate columns into count, mean, median, sum, std, min, and max. Return a
    new (Geo)DataFrame with a single row containing all aggregated values.
    Specify whether the geometries of the GeoDataFrame are to be aggregated.
    All geometries are merged union-like.

    The return data type will always be the same as the input data type.

    :param df: The (Geo)DataFrame to be analysed
    :param var_names: Variables to be aggregated ('None' uses all aggregatable columns)
    :param aggregate_geometry: Aggregate (union like) the geometry and add it
        to the resulting GeoDataFrame
    :param monitor: Monitor for progress bar
    :return: returns either DataFrame or GeoDataFrame. Keeps input data type
    """
    vns = VarNamesLike.convert(var_names)

    df_is_geo = isinstance(df, gpd.GeoDataFrame)
    aggregations = ["count", "mean", "median", "sum", "std", "min", "max"]

    # Check var names integrity (aggregatable, exists in data frame)
    types_accepted_for_agg = ['float64', 'int64', 'bool']
    agg_columns = list(df.select_dtypes(include=types_accepted_for_agg).columns)

    if df_is_geo:
        try:
            df['geometry']
        except KeyError as e:
            raise ValidationError('Variable geometry not in GEO data frame!') from e
        agg_columns.append('geometry')

    columns = list(df.columns)

    if vns is None:
        vns = agg_columns

    diff = list(set(vns) - set(columns))
    if len(diff) > 0:
        raise ValidationError('Variable ' + ','.join(diff) + ' not in data frame!')

    diff = list(set(vns) - set(agg_columns))
    if len(diff) > 0:
        raise ValidationError('Variable(s) ' + ','.join(diff) + ' not aggregatable!')

    # Aggregate columns
    if vns is None:
        df_buff = df.select_dtypes(include=types_accepted_for_agg).agg(aggregations)
    else:
        df_buff = df[vns].select_dtypes(
            include=types_accepted_for_agg
        ).agg(aggregations)

    res = {}
    for n in df_buff.columns:
        for a in aggregations:
            val = df_buff[n][a]
            h = n + '_' + a
            res[h] = [val]

    df_agg = pd.DataFrame(res)

    # Aggregate (union) geometry if GeoDataFrame
    if df_is_geo and aggregate_geometry:
        total_work = 100
        num_work_rows = 1 + len(df) // total_work
        with monitor.starting('Aggregating geometry: ', total_work):
            multi_polygon = shapely.geometry.MultiPolygon()
            i = 0
            for rec in df.geometry:
                if monitor.is_cancelled():
                    break
                # noinspection PyBroadException
                try:
                    multi_polygon = multi_polygon.union(other=rec)
                except Exception:
                    pass

                if i % num_work_rows == 0:
                    monitor.progress(work=1)
                i += 1

        df_agg = gpd.GeoDataFrame(df_agg, geometry=[multi_polygon], crs=df.crs)

    return df_agg


@op(tags=['filter'], version='1.0')
@op_input('df', data_type=DataFrameLike)
@op_input('var', value_set_source='df', data_type=VarName)
def data_frame_min(df: DataFrameLike.TYPE, var: VarName.TYPE) -> pd.DataFrame:
    """
    Select the first record of a data frame for which the given variable value is
    minimal.

    :param df: The data frame or dataset.
    :param var: The variable.
    :return: A new, one-record data frame.
    """
    data_frame = DataFrameLike.convert(df)
    var_name = VarName.convert(var)
    row_index = data_frame[var_name].idxmin()
    row_frame = data_frame.loc[[row_index]]
    return _maybe_convert_to_geo_data_frame(data_frame, row_frame)


@op(tags=['filter'], version='1.0')
@op_input('df', data_type=DataFrameLike)
@op_input('var', value_set_source='df', data_type=VarName)
def data_frame_max(df: DataFrameLike.TYPE, var: VarName.TYPE) -> pd.DataFrame:
    """
    Select the first record of a data frame for which the given variable value is
    maximal.

    :param df: The data frame or dataset.
    :param var: The variable.
    :return: A new, one-record data frame.
    """
    data_frame = DataFrameLike.convert(df)
    var_name = VarName.convert(var)
    row_index = data_frame[var_name].idxmax()
    row_frame = data_frame.loc[[row_index]]
    return _maybe_convert_to_geo_data_frame(data_frame, row_frame)


@op(tags=['filter'], version='1.0')
@op_input('df', data_type=DataFrameLike)
@op_input('query_expr')
def query(df: DataFrameLike.TYPE, query_expr: str) -> pd.DataFrame:
    """
    Select records from the given data frame where the given conditional query
    expression evaluates to "True".

    If the data frame *df* contains a geometry column (a ``GeoDataFrame`` object),
    then the query expression *query_expr* can also contain geometric relationship
    tests, for example the expression
    ``"population > 100000 and @within('-10, 34, 20, 60')"`` could be used on a data
    frame with the *population* and a *geometry* column to query for larger cities in
    West-Europe.

    The geometric relationship tests are
    * ``@almost_equals(geom)`` - does a feature's geometry almost equal
    the given ``geom``;
    * ``@contains(geom)`` - does a feature's geometry contain the given ``geom``;
    * ``@crosses(geom)`` - does a feature's geometry cross the given ``geom``;
    * ``@disjoint(geom)`` - does a feature's geometry not at all intersect
    the given ``geom``;
    * ``@intersects(geom)`` - does a feature's geometry intersect with given ``geom``;
    * ``@touches(geom)`` - does a feature's geometry have a point in common with
    given ``geom`` but does not intersect it;
    * ``@within(geom)`` - is a feature's geometry contained within given ``geom``.

    The *geom* argument may be a point ``"<lon>, <lat>"`` text string,
    a bounding box ``"<lon1>, <lat1>, <lon2>, <lat2>"`` text, or any valid geometry WKT.

    :param df: The data frame or dataset.
    :param query_expr: The conditional query expression.
    :return: A new data frame.
    """
    data_frame = DataFrameLike.convert(df)

    local_dict = dict(from_wkt=GeometryLike.convert)
    if hasattr(data_frame, 'geometry') \
            and isinstance(data_frame.geometry, gpd.GeoSeries):

        source_crs = dict(init='epsg:4326')
        try:
            target_crs = data_frame.crs.to_dict() or source_crs
        except AttributeError:
            target_crs = source_crs
        reprojection_func = _get_reprojection_func(source_crs, target_crs)

        def _almost_equals(geometry: GeometryLike):
            return _data_frame_geometry_op(data_frame.geometry.geom_almost_equals,
                                           geometry,
                                           reprojection_func)

        def _contains(geometry: GeometryLike):
            return _data_frame_geometry_op(data_frame.geometry.contains,
                                           geometry,
                                           reprojection_func)

        def _crosses(geometry: GeometryLike):
            return _data_frame_geometry_op(data_frame.geometry.crosses,
                                           geometry,
                                           reprojection_func)

        def _disjoint(geometry: GeometryLike):
            return _data_frame_geometry_op(data_frame.geometry.disjoint,
                                           geometry,
                                           reprojection_func)

        def _intersects(geometry: GeometryLike):
            return _data_frame_geometry_op(data_frame.geometry.intersects,
                                           geometry,
                                           reprojection_func)

        def _touches(geometry: GeometryLike):
            return _data_frame_geometry_op(data_frame.geometry.touches,
                                           geometry,
                                           reprojection_func)

        def _within(geometry: GeometryLike):
            return _data_frame_geometry_op(data_frame.geometry.within,
                                           geometry,
                                           reprojection_func)

        local_dict['almost_equals'] = _almost_equals
        local_dict['contains'] = _contains
        local_dict['crosses'] = _crosses
        local_dict['disjoint'] = _disjoint
        local_dict['intersects'] = _intersects
        local_dict['touches'] = _touches
        local_dict['within'] = _within

    # noinspection PyTypeChecker
    subset_data_frame = data_frame.query(query_expr,
                                         local_dict=local_dict,
                                         global_dict={})

    return _maybe_convert_to_geo_data_frame(data_frame, subset_data_frame)


def _data_frame_geometry_op(instance_method,
                            geometry: GeometryLike,
                            reprojection_func: ReprojectionFunc) -> bool:
    geometry = GeometryLike.convert(geometry)
    if geometry is None:
        return False
    geometry = _transform_coordinates(geometry, reprojection_func)
    if geometry is None:
        return False
    return instance_method(geometry)


@op(tags=['filter'], version='1.0')
@op_input('df', data_type=DataFrameLike)
def to_dataset(df: DataFrameLike.TYPE) -> xr.Dataset:
    """
    Convert a dataframe to a dataset.

    :param df: The dataframe.
    :return: A dataset created from the dataframe.
    """
    return DatasetLike.convert(df)


@op(tags=['filter'], version='1.0')
@op_input('ds', data_type=DataFrameLike)
def to_dataframe(ds: DatasetLike.TYPE) -> pd.DataFrame:
    """
    Convert a dataset into a dataframe.

    :param ds: The dataset.
    :return: A dataframe created from the geodataframe.
    """
    return DataFrameLike.convert(ds)


@op(tags=['filter'], version='1.0')
@op_input('df', data_type=DataFrameLike)
@op_input('geometry', data_type=VarName)
@op_input('spatial_coords', data_type=VarNamesLike)
@op_input('crs', data_type=Literal, default_value="EPSG:4362")
def as_geodataframe(
        df: DataFrameLike.TYPE,
        geometry: VarName.TYPE,
        spatial_coords: VarNamesLike.TYPE,
        crs: Literal.TYPE
) -> GeoDataFrame:
    """
    Converts a dataframe into a geodataframe, if possible. The conversion is possible
    if the dataframe contains either a geometry column or columns containing spatial
    coordinates.
    This operation supports specifying which column contains geometries or spatial
    coordinates. If nor value is given, it will be attempted to use a column
    named "geometry"; if this is not given, common name-pair-combinations will be tried.

    :param df: The dataframe.
    :param geometry: The name of the column holding the geometry.
    :param spatial_coords: The names of the columns holding spatial coordinates.
        Longitudinal coordinates come first, Latitudinal second
        Will not be considered when ``geometry`` is given.
    :param crs: The spatial reference system of the geometry.
        Defaults to ``EPSG:4362`` / ``WGS84``.
    :return: A geodataframe created from the dataframe.
    """
    if geometry is not None:
        return gpd.GeoDataFrame(df, geometry=geometry, crs=crs)
    elif spatial_coords is not None:
        if not isinstance(spatial_coords, list) or not len(spatial_coords) == 2:
            raise ValidationError(
                f"Invalid parameter 'spatial_coords'. "
                f"Must be list of exactly two names, was '{spatial_coords}'."
            )
        return gpd.GeoDataFrame(
            df,
            geometry=gpd.points_from_xy(df[spatial_coords[0]], df[spatial_coords[1]]),
            crs=crs
        )
    elif "geometry" in df:
        return gpd.GeoDataFrame(df, geometry=geometry, crs=crs)
    combination_sets = [
        ["lon", "lat"],
        ["longitude", "latitude"],
        ["x", "y"]
    ]
    for combination_set in combination_sets:
        if combination_set[0] in df and combination_set[1] in df:
            return gpd.GeoDataFrame(
                df,
                geometry=gpd.points_from_xy(combination_set[0], combination_set[1]),
                crs=crs
            )
    raise ValueError("Could not convert dataframe to geodataframe")


@op(tags=['filter'], version='1.0')
@op_input('gdf', data_type=GeoDataFrame)
@op_input('ds', data_type=DatasetLike)
@op_input('var_names', data_type=VarNamesLike)
@op_input('interpolation', value_set=['nearest', 'linear'], default_value='nearest')
@op_input('time_name', data_type=Literal)
@op_input('var_prefix', data_type=Literal)
def add_dataset_values_to_geodataframe(
        gdf: GeoDataFrame,
        ds: DatasetLike.TYPE,
        var_names: VarNamesLike.TYPE,
        interpolation: str = "nearest",
        time_name: Literal = None,
        var_prefix: Literal = None
) -> GeoDataFrame:
    """
    For each geometry in the input geodataframe, the value of the variables
    in the dataset will be determined. If the geometry is not a point,
    its centroid value will be used as point of reference.
    Note that this method also matches time.

    :param gdf: The geodataframe.
    :param ds: The dataset to read variables from.
    :param var_names: Names of variables for which to get values
    :param interpolation: Type of interpolation between dataset pixel centers used
        to derive value for dataframe points.
    :param time_name: Name of time column in the geodataframe. If not given, it is
        either expected to be ``time`` or non-existent.
        :param var_names: Names of variables for which to get values
    :param var_prefix: Prefix the columns of the extracted variables will carry
        in the geodataframe. Default is None.
    :return: The geodataframe, enriched by values from the dataset.
    """
    gdfc = gdf.copy()

    def maybe_convert_geometry(gdf):
        return shapely.centroid(gdf["geometry"])

    gdfc['geometry'] = gdfc.apply(maybe_convert_geometry, axis=1)
    ds_gm = GridMapping.from_dataset(ds)
    ds_xy_names = ds_gm.xy_var_names

    gdfc[ds_xy_names[0]] = gdfc['geometry'].x
    gdfc[ds_xy_names[1]] = gdfc['geometry'].y
    if time_name is not None:
        gdfc['time'] = pd.to_datetime(gdfc[time_name])

    ref_name_pattern = str(var_prefix) + "_{name}" if var_prefix else "{name}"

    cvs = get_cube_values_for_points(
        ds, points=gdf,
        var_names=var_names,
        ref_name_pattern=ref_name_pattern,
        method=interpolation
    )

    for var_name in var_names:
        gdf[var_name] = cvs[var_name]

    return gdf


@op(tags=['filter'], version='1.0')
@op_input('gdf', data_type=DataFrameLike)
@op_input('region_op', data_type=str, value_set=_REGION_MODES)
@op_input('region', data_type=PolygonLike)
@op_input('var_names', value_set_source='gdf', data_type=VarNamesLike)
def data_frame_subset(gdf: gpd.GeoDataFrame,
                      region_op: bool = 'intersects',
                      region: PolygonLike.TYPE = None,
                      var_names: VarNamesLike.TYPE = None) -> gpd.GeoDataFrame:
    """
    Create a GeoDataFrame subset from given variables (data frame columns)
    and/or region.

    :param gdf: A GeoDataFrame.
    :param region_op: The geometric operation to be performed if *region* is given.
    :param region: A region polygon used to filter rows.
    :param var_names: The variables (columns) to select.
    :return: A GeoDataFrame subset.
    """

    region = PolygonLike.convert(region)

    var_names = VarNamesLike.convert(var_names)

    if not var_names and not region:
        return gdf

    if var_names:
        if 'geometry' not in var_names:
            var_names = ['geometry'] + var_names
        gdf = gdf[var_names]

    if region and region_op:
        geom_str = PolygonLike.format(region)
        gdf = query(gdf, f'@{region_op}("{geom_str}")')

    return gdf


@op(tags=['filter'], version='1.0')
@op_input('gdf', data_type=DataFrameLike)
@op_input('location', data_type=GeometryLike)
@op_input('max_results')
@op_input('max_dist')
@op_input('dist_col_name')
def find_closest(gdf: gpd.GeoDataFrame,
                 location: GeometryLike.TYPE,
                 max_results: int = 1,
                 max_dist: float = 180,
                 dist_col_name: str = 'distance',
                 monitor: Monitor = Monitor.NONE) -> gpd.GeoDataFrame:
    """
    Find the *max_results* records closest to given *location* in the given
    GeoDataFrame *gdf*.
    Return a new GeoDataFrame containing the closest records.

    If *dist_col_name* is given, store the actual distances in this column.

    Distances are great-circle distances measured in degrees from a representative
    center of the given *location* geometry to the representative centres of each
    geometry in the *gdf*.

    :param gdf: The GeoDataFrame.
    :param location: A location given as arbitrary geometry.
    :param max_results: Maximum number of results.
    :param max_dist: Ignore records whose distance is greater than this value in
        degrees.
    :param dist_col_name: Optional name of a new column that will store the actual
        distances.
    :param monitor: A progress monitor.
    :return: A new GeoDataFrame containing the closest records.
    """
    location = GeometryLike.convert(location)
    location_point = location.representative_point()

    target_crs = dict(init='epsg:4326')
    try:
        source_crs = gdf.crs or target_crs
    except AttributeError:
        source_crs = target_crs
    reprojection_func = _get_reprojection_func(source_crs, target_crs)

    try:
        geometries = gdf.geometry
    except AttributeError as e:
        raise ValidationError('Missing default geometry column in data frame.') from e

    num_rows = len(geometries)
    indexes = list()

    # PERF: Note, this operation may be optimized by computing the great-circle
    # distances using numpy array math!

    total_work = 100
    num_work_rows = 1 + num_rows // total_work
    with monitor.starting('Finding closest records', total_work):
        for i in range(num_rows):
            geometry = geometries.iloc[i]
            if geometry is not None:
                # noinspection PyBroadException
                try:
                    representative_point = geometry.representative_point()
                except BaseException:
                    # For some geometries shapely.representative_point() raises
                    # AttributeError or ValueError.
                    # E.g. features that span the poles will raise ValueError.
                    # The quick and dirty solution here is to catch such exceptions
                    # and ignore them.
                    representative_point = None
                if representative_point is not None:
                    representative_point = _transform_coordinates(representative_point,
                                                                  reprojection_func)
                    if representative_point is not None:
                        # noinspection PyTypeChecker
                        dist = great_circle_distance(
                            location_point, representative_point
                        )
                        if dist <= max_dist:
                            indexes.append((i, dist))
            if i % num_work_rows == 0:
                monitor.progress(work=1)

    indexes = sorted(indexes, key=lambda item: item[1])
    num_results = min(max_results, len(indexes))
    indexes, distances = zip(*indexes[0:num_results])

    new_gdf = gdf.iloc[list(indexes)]
    if not isinstance(new_gdf, gpd.GeoDataFrame):
        new_gdf = gpd.GeoDataFrame(new_gdf, crs=source_crs)

    if dist_col_name:
        new_gdf[dist_col_name] = da.array(distances)

    return new_gdf


def great_circle_distance(
        p1: shapely.geometry.Point, p2: shapely.geometry.Point
) -> float:
    """
    Compute great-circle distance on a Sphere in degrees.

    See https://en.wikipedia.org/wiki/Great-circle_distance.

    :param p1: First point (lon, lat) in degrees
    :param p2: Second point (lon, lat) in degrees
    :return: Great-circle distance in degree
    """

    lam1, phi1 = p1.x, p1.y
    lam2, phi2 = p2.x, p2.y
    lambda_diff = abs(lam2 - lam1)

    if lambda_diff > 180.:
        lambda_diff = 360. - lambda_diff

    phi1 *= _DEG2RAD
    phi2 *= _DEG2RAD
    lambda_diff *= _DEG2RAD

    sin_phi1 = math.sin(phi1)
    cos_phi1 = math.cos(phi1)
    sin_phi2 = math.sin(phi2)
    cos_phi2 = math.cos(phi2)
    sin_dlam = math.sin(lambda_diff)
    cos_dlam = math.cos(lambda_diff)

    dx = cos_phi2 * sin_dlam
    dy = cos_phi1 * sin_phi2 - sin_phi1 * cos_phi2 * cos_dlam

    y = math.sqrt(dx * dx + dy * dy)
    x = sin_phi1 * sin_phi2 + cos_phi1 * cos_phi2 * cos_dlam

    return math.atan2(y, x) / _DEG2RAD


# Transforms a geometry that is used in an operator for e.g. feature selection purposes.
# It assures that both use an identical 'crs' (projection)
def _transform_coordinates(geometry: shapely.geometry.base.BaseGeometry,
                           reprojection_func: ReprojectionFunc):
    if reprojection_func is not None:
        # noinspection PyBroadException
        try:
            return shapely.ops.transform(reprojection_func, geometry)
        except BaseException as e:
            warnings.warn(f'coordinate transformation failed: {e}')
            return None
    else:
        return geometry


# Transforms a geometry that is used in an operator for e.g. feature selection purposes.
# It assures that both use an identical 'crs' (projection)
def _get_reprojection_func(source_crs: Dict[str, Any],
                           target_crs: Dict[str, Any]) -> Optional[ReprojectionFunc]:
    if source_crs and target_crs and source_crs != target_crs:
        # noinspection PyBroadException,PyTypeChecker
        return functools.partial(
            pyproj.transform,
            pyproj.Proj(**source_crs),  # source coordinate system
            pyproj.Proj(**target_crs)  # target coordinate system
        )
    else:
        return None


def _maybe_convert_to_geo_data_frame(data_frame: pd.DataFrame,
                                     data_frame_2: pd.DataFrame) -> pd.DataFrame:
    if isinstance(data_frame, gpd.GeoDataFrame) \
            and not isinstance(data_frame_2, gpd.GeoDataFrame):
        return gpd.GeoDataFrame(data_frame_2, crs=data_frame.crs)
    else:
        return data_frame_2
